# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
# created 2019-02-15
import pytest

from ezdxf.entities.image import ImageDef
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text

IMAGEDEF = """0
IMAGEDEF
  5
0
102
{ACAD_REACTORS
330
0
102
}
330
0
100
AcDbRasterImageDef
 90
  0
  1
path/filename.jpg
 10
640
 20
480
 11
0.01
 21
0.01
280
  1
281
  0
"""


@pytest.fixture
def entity():
    return ImageDef.from_text(IMAGEDEF)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "IMAGEDEF" in ENTITY_CLASSES


def test_default_init():
    entity = ImageDef()
    assert entity.dxftype() == "IMAGEDEF"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = ImageDef.new(
        handle="ABBA", owner="0", dxfattribs={"filename": "image.jpg"}
    )
    assert entity.dxf.class_version == 0
    assert entity.dxf.filename == "image.jpg"
    assert entity.dxf.image_size is None
    assert entity.dxf.pixel_size == (0.01, 0.01)
    assert entity.dxf.loaded == 1
    assert entity.dxf.resolution_units == 0


def test_load_from_text(entity):
    assert entity.dxf.class_version == 0
    assert entity.dxf.filename == "path/filename.jpg"
    assert entity.dxf.image_size == (640, 480)
    assert entity.dxf.pixel_size == (0.01, 0.01)
    assert entity.dxf.loaded == 1
    assert entity.dxf.resolution_units == 0


def test_write_dxf():
    entity = ImageDef.from_text(IMAGEDEF)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(IMAGEDEF)
    assert result == expected



# Copyright (c) 2018-2021 Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.entities.dimension import Dimension
from ezdxf.protocols import SupportsVirtualEntities, query_virtual_entities


@pytest.fixture(scope="module")
def dxf2000():
    return ezdxf.new("R2000", setup="all")


@pytest.fixture(scope="module")
def dxf2007():
    return ezdxf.new("R2007", setup="all")


def test_dimstyle_standard_exist(dxf2000):
    assert "EZDXF" in dxf2000.dimstyles


def test_rotated_dimline(dxf2000):
    msp = dxf2000.modelspace()
    dxfattribs = {"dimtype": Dimension.LINEAR}
    dimline = msp.new_entity("DIMENSION", dxfattribs)
    assert dimline.dimtype == Dimension.LINEAR
    assert dimline.dxf.defpoint == (0, 0, 0)
    assert dimline.dxf.insert == (0, 0, 0)
    assert dimline.dxf.defpoint2 == (0, 0, 0)
    assert dimline.dxf.defpoint3 == (0, 0, 0)
    assert dimline.dxf.angle == 0.0
    assert dimline.dxf.oblique_angle == 0.0


def test_aligned_dimline(dxf2000):
    msp = dxf2000.modelspace()
    dxfattribs = {"dimtype": Dimension.ALIGNED}
    dimline = msp.new_entity("DIMENSION", dxfattribs)
    assert dimline.dimtype == Dimension.ALIGNED
    assert dimline.dxf.defpoint == (0, 0, 0)
    assert dimline.dxf.insert == (0, 0, 0)
    assert dimline.dxf.defpoint2 == (0, 0, 0)
    assert dimline.dxf.defpoint3 == (0, 0, 0)
    assert dimline.dxf.angle == 0.0
    assert dimline.dxf.oblique_angle == 0.0


def test_angular_dimline(dxf2000):
    msp = dxf2000.modelspace()
    dxfattribs = {"dimtype": Dimension.ANGULAR}
    dimline = msp.new_entity("DIMENSION", dxfattribs)
    assert dimline.dimtype == Dimension.ANGULAR
    assert dimline.dxf.defpoint == (0, 0, 0)
    assert dimline.dxf.defpoint2 == (0, 0, 0)
    assert dimline.dxf.defpoint3 == (0, 0, 0)
    assert dimline.dxf.defpoint4 == (0, 0, 0)
    assert dimline.dxf.defpoint5 == (0, 0, 0)


def test_angular_3p_dimline(dxf2000):
    msp = dxf2000.modelspace()
    dxfattribs = {"dimtype": Dimension.ANGULAR_3P}
    dimline = msp.new_entity("DIMENSION", dxfattribs)
    assert dimline.dimtype == Dimension.ANGULAR_3P


def test_radius_dimline(dxf2000):
    msp = dxf2000.modelspace()
    dxfattribs = {"dimtype": Dimension.RADIUS}
    dimline = msp.new_entity("DIMENSION", dxfattribs)
    assert dimline.dimtype == Dimension.RADIUS
    assert dimline.dxf.defpoint == (0, 0, 0)
    assert dimline.dxf.defpoint4 == (0, 0, 0)
    assert dimline.is_supported_dxf_attrib("leader_length")


def test_diameter_dimline(dxf2000):
    msp = dxf2000.modelspace()
    dxfattribs = {"dimtype": Dimension.DIAMETER}
    dimline = msp.new_entity("DIMENSION", dxfattribs)
    assert dimline.dimtype == Dimension.DIAMETER
    assert dimline.dxf.defpoint == (0, 0, 0)
    assert dimline.dxf.defpoint4 == (0, 0, 0)
    assert dimline.is_supported_dxf_attrib("leader_length")


def test_ordinate_dimline(dxf2000):
    msp = dxf2000.modelspace()
    dxfattribs = {"dimtype": Dimension.ORDINATE}
    dimline = msp.new_entity("DIMENSION", dxfattribs)
    assert dimline.dimtype == Dimension.ORDINATE
    assert dimline.dxf.defpoint == (0, 0, 0)
    assert dimline.dxf.defpoint2 == (0, 0, 0)
    assert dimline.dxf.defpoint3 == (0, 0, 0)


def add_linear_dimension(doc):
    msp = doc.modelspace()
    override = msp.add_linear_dim(
        base=(3, 2, 0),
        p1=(0, 0, 0),
        p2=(3, 0, 0),
    )
    override.render()
    return override.dimension


def test_add_horizontal_dimline(dxf2000):
    dimline = add_linear_dimension(dxf2000)
    assert dimline.dxf.dimstyle == "EZDXF"
    block_name = dimline.dxf.geometry
    assert block_name.startswith("*D")

    block = dimline.get_geometry_block()
    assert len(list(block.query("MTEXT"))) == 1
    assert len(list(block.query("INSERT"))) == 2
    assert (
        len(list(block.query("LINE"))) == 3
    )  # dimension line + 2 extension lines
    assert len(list(block.query("POINT"))) == 3  # def points


def test_virtual_entities_and_explode(dxf2000):
    dimline = add_linear_dimension(dxf2000)

    parts = list(dimline.virtual_entities())
    assert len(parts) == 9
    geometry = dimline.dxf.geometry
    parts = dimline.explode()
    assert len(list(parts.query("MTEXT"))) == 1
    assert len(list(parts.query("INSERT"))) == 2
    assert (
        len(list(parts.query("LINE"))) == 3
    )  # dimension line + 2 extension lines
    assert len(list(parts.query("POINT"))) == 3  # def points
    assert dimline.is_alive is False
    assert (
        geometry in dxf2000.blocks
    ), "Do not destroy anonymous block, may be used by block references."


def test_transformation_of_associated_anonymous_geometry_block(dxf2000):
    dimline = add_linear_dimension(dxf2000)
    dx, dy = 1, 1
    block = dimline.get_geometry_block()
    original_points = [e.dxf.location for e in block if e.dxftype() == "POINT"]
    dimline.translate(dx, dy, 0)
    transformed_points = [
        e.dxf.location for e in block if e.dxftype() == "POINT"
    ]
    for o, t in zip(original_points, transformed_points):
        assert t.isclose(o + (dx, dy))


def test_copy_dimension_with_geometry_block(dxf2000):
    dimline = add_linear_dimension(dxf2000)
    vcopy = dimline.copy()
    assert vcopy.virtual_block_content is not None
    assert vcopy.dxf.hasattr("geometry") is False
    block = dimline.get_geometry_block()
    assert len(vcopy.virtual_block_content) == len(block)
    for copy, original in zip(vcopy.virtual_block_content, block):
        assert copy.is_virtual is True
        assert original.is_virtual is False
        assert copy.dxftype() == original.dxftype()


def test_destroy_virtual_dimension_copy(dxf2000):
    dimline = add_linear_dimension(dxf2000)
    dimline.destroy()
    assert hasattr(dimline, "virtual_block_content") is False


def test_transform_virtual_geometry_block(dxf2000):
    original = add_linear_dimension(dxf2000)

    original_points = [
        e.dxf.location
        for e in original.virtual_entities()
        if e.dxftype() == "POINT"
    ]
    vcopy = original.copy()
    dx, dy = 1, 1
    vcopy.translate(dx, dy, 0)
    transformed_points = [
        e.dxf.location
        for e in vcopy.virtual_entities()
        if e.dxftype() == "POINT"
    ]
    assert len(transformed_points) == len(original_points)
    for o, t in zip(original_points, transformed_points):
        assert t.isclose(o + (dx, dy))


def test_add_virtual_dimension_copy_to_layout(dxf2000):
    dimline = add_linear_dimension(dxf2000)
    vcopy = dimline.copy()
    msp = dxf2000.modelspace()
    content = vcopy.virtual_block_content
    msp.add_entity(vcopy)
    assert vcopy.is_virtual is False
    assert vcopy.virtual_block_content is None
    assert all(
        not e.is_virtual for e in content
    ), "all entities should be non virtual"
    db = dxf2000.entitydb
    assert all(
        e.dxf.handle in db for e in content
    ), "all entities should be stored in the entity database"
    assert all(
        e.doc is dxf2000 for e in content
    ), "all entities should be bound to the document"


def test_supports_virtual_entities_protocol(dxf2000):
    dimline = add_linear_dimension(dxf2000)
    assert isinstance(dimline, SupportsVirtualEntities) is True
    assert len(query_virtual_entities(dimline)) > 0


def test_dimstyle_override(dxf2000):
    msp = dxf2000.modelspace()
    dimstyle = msp.add_linear_dim(
        base=(3, 2, 0),
        p1=(0, 0, 0),
        p2=(3, 0, 0),
        dxfattribs={
            "dimstyle": "EZDXF",
        },
    )
    dimline = dimstyle.dimension
    assert dimline.dxf.dimstyle == "EZDXF"
    if "TEST" not in dxf2000.styles:  # text style must exists
        dxf2000.styles.new("TEST")

    preset = {
        "dimtxsty": "TEST",
        "dimexe": 0.777,
    }
    dimstyle.update(preset)
    assert dimstyle["dimtxsty"] == "TEST"
    assert dimstyle["dimexe"] == 0.777

    assert dimstyle["invalid"] is None
    dimstyle.update({"invalid": 7})
    # ezdxf 0.10 and later uses internally only resource names not handles for dim style attributes
    # unknown attributes are ignored
    dstyle_orig = dimstyle.get_dstyle_dict()
    assert len(dstyle_orig) == 0

    dimstyle.commit()
    # ezdxf 0.10 and later uses internally only resource names not handles for dim style attributes
    dstyle = dimstyle.get_dstyle_dict()

    assert dstyle["dimexe"] == 0.777

    # handle attributes not available, just stored transparent in XDATA
    assert "dimtxsty_handle" not in dstyle

    assert dstyle["dimtxsty"] == "TEST"


def test_linetype_override_R2000(dxf2000):
    msp = dxf2000.modelspace()
    preset = {
        "dimltype": "DOT",
        "dimltex1": "DOT2",
        "dimltex2": "DOTX2",
    }
    dimstyle = msp.add_linear_dim(
        base=(3, 2, 0),
        p1=(0, 0, 0),
        p2=(3, 0, 0),
        dimstyle="EZDXF",
        override=preset,
    )
    assert dimstyle["dimltype"] == "DOT"
    assert dimstyle["dimltex1"] == "DOT2"
    assert dimstyle["dimltex2"] == "DOTX2"

    dimstyle.commit()
    # ezdxf 0.10 and later uses internally only resource names not handles for dim style attributes
    dstyle = dimstyle.get_dstyle_dict()

    # handle attributes not available, just stored transparent in XDATA
    assert "dimltype_handle" not in dstyle
    assert "dimltex1_handle" not in dstyle
    assert "dimltex2_handle" not in dstyle

    # line type not supported by DXF R2000
    assert "dimltype" not in dstyle
    assert "dimltex1" not in dstyle
    assert "dimltex2" not in dstyle


def test_linetype_override_R2007(dxf2007):
    msp = dxf2007.modelspace()
    preset = {
        "dimltype": "DOT",
        "dimltex1": "DOT2",
        "dimltex2": "DOTX2",
    }
    dimstyle = msp.add_linear_dim(
        base=(3, 2, 0),
        p1=(0, 0, 0),
        p2=(3, 0, 0),
        dimstyle="EZDXF",
        override=preset,
    )
    assert dimstyle["dimltype"] == "DOT"
    assert dimstyle["dimltex1"] == "DOT2"
    assert dimstyle["dimltex2"] == "DOTX2"

    dimstyle.commit()
    # ezdxf 0.10 and later uses internally only resource names not handles for dim style attributes
    dstyle = dimstyle.get_dstyle_dict()

    # handle attributes not available, just stored transparent in XDATA
    assert "dimltype_handle" not in dstyle
    assert "dimltex1_handle" not in dstyle
    assert "dimltex2_handle" not in dstyle

    assert dstyle["dimltype"] == "DOT"
    assert dstyle["dimltex1"] == "DOT2"
    assert dstyle["dimltex2"] == "DOTX2"


def test_dimstyle_override_arrows(dxf2000):
    msp = dxf2000.modelspace()
    arrows = ezdxf.ARROWS
    blocks = dxf2000.blocks

    arrows.create_block(blocks, arrows.dot_blank)
    arrows.create_block(blocks, arrows.box)
    arrows.create_block(blocks, arrows.closed)
    arrows.create_block(blocks, arrows.closed_filled)

    preset = {
        "dimblk": arrows.dot_blank,
        "dimblk1": arrows.box,
        "dimblk2": arrows.closed,
        "dimldrblk": arrows.closed_filled,  # virtual attribute
    }
    dimstyle = msp.add_linear_dim(
        base=(3, 2, 0),
        p1=(0, 0, 0),
        p2=(3, 0, 0),
        dimstyle="EZDXF",
        override=preset,
    )
    # still as block names stored
    assert dimstyle["dimblk"] == arrows.dot_blank
    assert dimstyle["dimblk1"] == arrows.box
    assert dimstyle["dimblk2"] == arrows.closed
    assert dimstyle["dimldrblk"] == arrows.closed_filled

    dstyle_orig = dimstyle.get_dstyle_dict()
    assert len(dstyle_orig) == 0

    dimstyle.commit()
    # ezdxf 0.10 and later uses internally only resource names not handles for dim style attributes
    dstyle = dimstyle.get_dstyle_dict()

    # handle attributes not available, just stored transparent in XDATA
    assert "dimblk_handle" not in dstyle
    assert "dimblk1_handle" not in dstyle
    assert "dimblk2_handle" not in dstyle
    assert "dimldrblk_handle" not in dstyle

    assert dstyle["dimblk"] == arrows.dot_blank
    assert dstyle["dimblk1"] == arrows.box
    assert dstyle["dimblk2"] == arrows.closed
    assert dstyle["dimldrblk"] == ""  # special handle for closed filled

    dimstyle.set_arrows(
        blk=arrows.closed,
        blk1=arrows.dot_blank,
        blk2=arrows.box,
        ldrblk=arrows.dot_small,
    )
    assert dimstyle["dimblk"] == arrows.closed
    assert dimstyle["dimblk1"] == arrows.dot_blank
    assert dimstyle["dimblk2"] == arrows.box
    assert dimstyle["dimldrblk"] == arrows.dot_small

    dimstyle.commit()
    # ezdxf 0.10 and later uses internally only resource names not handles for dim style attributes
    dstyle = dimstyle.get_dstyle_dict()
    assert dstyle["dimblk"] == arrows.closed
    assert dstyle["dimblk1"] == arrows.dot_blank
    assert dstyle["dimblk2"] == arrows.box
    # create acad arrows on demand
    assert dstyle["dimldrblk"] == arrows.dot_small



# Copyright (c) 2018-2019 Manfred Moitzi
# License: MIT License
import pytest
import os
import ezdxf
from ezdxf.lldxf.const import LATEST_DXF_VERSION


def test_lwpolyline(tmpdir):
    dwg = ezdxf.new(LATEST_DXF_VERSION)
    msp = dwg.modelspace()
    # point format = (x, y, [start_width, [end_width, [bulge]]])
    points = [
        (0, 0, 0, 0.05),
        (3, 0, 0.1, 0.2, -0.5),
        (6, 0, 0.1, 0.05),
        (9, 0),
    ]

    msp.add_lwpolyline(points)
    filename = str(tmpdir.join("lwpolyline.dxf"))
    try:
        dwg.saveas(filename)
    except ezdxf.DXFError as e:
        pytest.fail(
            "DXFError: {0} for DXF version {1}".format(str(e), dwg.dxfversion)
        )
    assert os.path.exists(filename)

    del dwg
    dwg = ezdxf.readfile(filename)
    msp = dwg.modelspace()
    lwpolyline = msp.query("LWPOLYLINE")[0]
    assert len(lwpolyline) == 4

    pts = lwpolyline.get_points()
    assert pts[0] == (0, 0, 0, 0.05, 0)
    assert pts[1] == (3, 0, 0.1, 0.2, -0.5)
    assert pts[2] == (6, 0, 0.1, 0.05, 0)
    assert pts[3] == (9, 0, 0, 0, 0)



# Copyright (c) 2013-2019, Manfred Moitzi
# License: MIT
import pytest
import os
import ezdxf
from ezdxf.addons.importer import Importer


def save_source_dwg(dwg, filename):
    if not os.path.exists(filename):
        dwg.saveas(filename)


def create_source_drawing(version):
    dwg = ezdxf.new(version)
    # complex shape linetype
    dwg.linetypes.new(
        "BORDER",
        dxfattribs={
            "description": "Border square ----[]-----[]----[]-----[]----[]--",
            "length": 1.45,
            "pattern": "A,.25,-.1,[132,ltypeshp.shx,x=-.1,s=.1],-.1,1",
        },
    )
    # complex text linetype
    dwg.linetypes.new(
        "GAS",
        dxfattribs={
            "description": "Gas ----GAS----GAS----GAS----GAS----GAS----GAS--",
            "length": 1,
            "pattern": 'A,.5,-.2,["GAS",STANDARD,S=.1,U=0.0,X=-0.1,Y=-.05],-.25',
        },
    )
    dwg.layers.new("Test", dxfattribs={"color": 17})
    dwg.layers.new("TestConflict", dxfattribs={"color": 18})
    msp = dwg.modelspace()
    msp.add_line((0, 0), (10, 0))
    msp.add_circle((0, 0), radius=5)
    msp.add_line((0, 1), (10, 1), dxfattribs={"linetype": "BORDER"})
    msp.add_line((0, 2), (10, 2), dxfattribs={"linetype": "GAS"})
    msp.add_blockref("TestBlock", insert=(0, 0))
    msp.add_blockref("ConflictBlock", insert=(0, 0))
    build_block(dwg, "TestBlock")
    build_block(dwg, "ConflictBlock")
    block = build_block(dwg, "RefToConflictBlock")
    block.add_blockref("ConflictBlock", insert=(0, 0))
    return dwg


def create_target_drawing(version):
    dwg = ezdxf.new(version)
    dwg.layers.new("TestConflict", dxfattribs={"color": 19})
    conflict_block = build_block(dwg, "ConflictBlock")
    conflict_block.add_circle((1, 1), radius=7)
    return dwg


def build_block(dwg, name):
    block = dwg.blocks.new(name=name)
    block.add_line((0, 0), (10, 0))
    block.add_circle((0, 0), radius=5)
    return block


SRC_DWG = {
    "R12": create_source_drawing("R12"),
    "R2000": create_source_drawing("R2000"),
}


@pytest.fixture(params=["R12", "R2000"])
def importer(request):
    version = request.param
    source = SRC_DWG[version]
    target = create_target_drawing(version)
    return Importer(source, target)


def test_import_simple_modelspace(importer):
    importer.import_modelspace()
    importer.finalize()

    source_entities = list(importer.source.modelspace())
    target_entities = list(importer.target.modelspace())
    assert len(source_entities) == len(target_entities)


def test_import_tables_without_conflict(importer):
    importer.import_table("layers")
    importer.finalize()

    layer = importer.target.layers.get("Test")
    assert layer.get_color() == 17


def test_import_tables_with_conflict_discard(importer):
    importer.import_table("layers", replace=False)
    importer.finalize()

    layer = importer.target.layers.get("TestConflict")
    assert layer.get_color() == 19


def test_import_tables_with_conflict_replace(importer):
    importer.import_table("layers", replace=True)
    importer.finalize()

    layer = importer.target.layers.get("TestConflict")
    assert layer.get_color() == 18


def test_import_block_without_conflict(importer):
    importer.import_block("TestBlock")
    importer.finalize()

    block = importer.target.blocks.get("TestBlock")
    block_entities = list(block)
    assert len(block_entities) == 2


def test_import_block_with_conflict_discard(importer):
    importer.import_blocks(["ConflictBlock"], rename=False)
    importer.finalize()

    block = importer.target.blocks.get("ConflictBlock")
    block_entities = list(block)
    assert len(block_entities) == 3


def test_import_block_with_conflict_rename(importer):
    importer.import_blocks(["ConflictBlock"], rename=True)
    importer.finalize()

    block = importer.target.blocks.get("ConflictBlock0")
    block_entities = list(block)
    assert len(block_entities) == 2


def test_import_block_with_conflict_rename_resolve_block_ref(importer):
    importer.import_blocks(["ConflictBlock"], rename=True)
    inserts = importer.source.query("INSERT")
    importer.import_entities(inserts)
    importer.finalize()

    msp = list(importer.target.modelspace())
    assert msp[1].dxf.name == "ConflictBlock0"


def test_import_block_with_conflict_rename_resolve_block_ref_inside_block_def(
    importer,
):
    src_blocks = [
        block.name
        for block in importer.source.blocks
        if not block.is_any_layout
    ]
    importer.import_blocks(src_blocks, rename=True)
    importer.finalize()

    block = importer.target.blocks.get("RefToConflictBlock")
    block_entities = list(block)
    block_ref_to_conflict_block = block_entities[2]
    assert block_ref_to_conflict_block.dxf.name == "ConflictBlock0"


@pytest.mark.parametrize(
    "name, style_name, font",
    [
        ("BORDER", "", "ltypeshp.shx"),  # shape file entry has no name
        ("GAS", "Standard", "txt"),  # text
    ],
)
def test_import_entities_using_complex_linetypes(
    importer, name, style_name, font
):
    lines = importer.source.query(f'LINE[linetype=="{name}"]')
    assert len(lines) == 1, "Required test LINE does not exist."

    importer.import_entities(lines)
    importer.finalize()

    assert name in importer.target.linetypes, f"Linetype {name } not imported."
    ltype = importer.target.linetypes.get(name)

    if importer.source.dxfversion == "AC1009":
        return  # complex line types are not supported by DXF R12

    style_handle = ltype.pattern_tags.get_style_handle()
    assert style_handle != "0", "Invalid shape file handle."

    style = importer.target.entitydb[style_handle]
    assert style.dxf.name == style_name
    assert style.dxf.font == font


def test_import_polyline():
    source = ezdxf.new()
    source.modelspace().add_polyline3d([(0, 0), (3, 0), (3, 3), (0, 3)])
    target = ezdxf.new()
    importer = Importer(source, target)
    importer.import_modelspace()
    tpoly = target.modelspace()[0]
    assert len(tpoly.vertices) == 4
    assert tpoly.seqend is not None
    assert tpoly.seqend.dxf.layer == tpoly.dxf.layer


def test_import_insert_with_attribs():
    source = ezdxf.new()
    source.blocks.new("Test")
    sinsert = source.modelspace().add_blockref("Test", insert=(0, 0))
    sinsert.add_attrib("A1", "text1")
    sinsert.add_attrib("A2", "text2")
    target = ezdxf.new()
    importer = Importer(source, target)
    importer.import_modelspace()
    tinsert = target.modelspace()[0]
    assert len(tinsert.attribs) == 2
    assert tinsert.seqend is not None
    assert tinsert.seqend.dxf.layer == tinsert.dxf.layer



# Copyright (c) 2019-2025 Manfred Moitzi
# License: MIT License
import pytest

import ezdxf
from ezdxf.entities.viewport import Viewport
from ezdxf.lldxf.extendedtags import ExtendedTags, DXFTag
from ezdxf.lldxf.const import DXF12, DXF2000
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from ezdxf.math import Z_AXIS

TEST_CLASS = Viewport
TEST_TYPE = "VIEWPORT"

ENTITY_R12 = """0
VIEWPORT
5
0
8
VIEWPORTS
10
0.0
20
0.0
30
0.0
40
1.0
41
1.0
68
1
1001
ACAD
1000
MVIEW
1002
{
1070
16
1010
0.0
1020
0.0
1030
0.0
1010
0.0
1020
0.0
1030
0.0
1040
0.0
1040
1.0
1040
0.0
1040
0.0
1040
50.0
1040
0.0
1040
0.0
1070
0
1070
100
1070
1
1070
3
1070
0
1070
0
1070
0
1070
0
1040
0.0
1040
0.0
1040
0.0
1040
0.1
1040
0.1
1040
0.1
1040
0.1
1070
0
1002
{
1002
}
1002
}
"""

ENTITY_R2000 = """0
VIEWPORT
5
0
330
0
100
AcDbEntity
67
1
8
VIEWPORTS
100
AcDbViewport
10
0.0
20
0.0
30
0.0
40
1.0
41
1.0
68
2
69
2
12
0.0
22
0.0
13
0.0
23
0.0
14
0.1
24
0.1
15
0.1
25
0.1
16
0.0
26
0.0
36
0.0
17
0.0
27
0.0
37
0.0
42
50.0
43
0.0
44
0.0
45
1.0
50
0.0
51
0.0
72
100
90
32864
1

281
0
71
0
74
0
110
0.0
120
0.0
130
0.0
111
1.0
121
0.0
131
0.0
112
0.0
122
1.0
132
0.0
79
0
146
0.0
"""


@pytest.fixture(params=[ENTITY_R12, ENTITY_R2000])
def entity(request):
    return TEST_CLASS.from_text(request.param)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert TEST_TYPE in ENTITY_CLASSES


def test_default_init():
    entity = TEST_CLASS()
    assert entity.dxftype() == TEST_TYPE


def test_default_new():
    entity = TEST_CLASS.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
            "view_direction_vector": (1, 2, 3),
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.view_direction_vector == (1, 2, 3)
    assert entity.dxf.view_direction_vector.x == 1, "is not Vec3 compatible"
    assert entity.dxf.view_direction_vector.y == 2, "is not Vec3 compatible"
    assert entity.dxf.view_direction_vector.z == 3, "is not Vec3 compatible"
    assert entity.dxf.view_target_point == (0, 0, 0)
    assert entity.dxf.view_twist_angle == 0
    assert entity.dxf.view_height == 1
    assert entity.dxf.view_center_point == (0, 0)
    assert entity.dxf.perspective_lens_length == 50
    assert entity.dxf.front_clip_plane_z_value == 0
    assert entity.dxf.back_clip_plane_z_value == 0
    assert entity.dxf.flags == 0
    assert entity.dxf.circle_zoom == 100
    assert entity.dxf.ucs_icon == 0
    assert entity.dxf.snap_angle == 0
    assert entity.dxf.snap_base_point == (0, 0)
    assert entity.dxf.snap_spacing == (10, 10)
    assert entity.dxf.grid_spacing == (10, 10)
    assert len(list(entity.frozen_layers)) == 0


def test_get_view_direction_vector_defaults_to_z_axis():
    vp = TEST_CLASS.new(dxfattribs={"view_direction_vector": (0, 0, 0)})
    assert vp.get_view_direction().isclose(Z_AXIS)


def test_get_view_direction_vector_is_normalized():
    vp = TEST_CLASS.new(dxfattribs={"view_direction_vector": (0, 0, 100)})
    assert vp.get_view_direction().isclose(Z_AXIS)


def test_load_from_text(entity):
    assert entity.dxf.layer == "VIEWPORTS"
    assert entity.dxf.center == (0, 0, 0)


def test_write_dxf_r2000():
    expected = basic_tags_from_text(ENTITY_R2000)
    viewport = TEST_CLASS.from_text(ENTITY_R2000)
    collector = TagCollector(dxfversion=DXF2000, optional=True)
    viewport.export_dxf(collector)
    # todo: tag checking
    assert len(collector.tags) == len(expected)

    collector2 = TagCollector(dxfversion=DXF2000, optional=False)
    viewport.export_dxf(collector2)
    assert collector.has_all_tags(collector2)


def test_write_dxf_r12():
    viewport = TEST_CLASS.from_text(ENTITY_R12)
    collector = TagCollector(dxfversion=DXF12, optional=True)
    viewport.export_dxf(collector)
    xdata = ExtendedTags(DXFTag(c, v) for c, v in collector.tags).xdata[0]
    assert xdata[0] == (1001, "ACAD")
    assert xdata[1] == (1000, "MVIEW")
    assert xdata[2] == (1002, "{")
    assert xdata[-1] == (1002, "}")


def test_post_load_hook_resolves_frozen_layer_handles_into_names():
    doc = ezdxf.new("R2000")
    l1 = doc.layers.new("Layer1")
    l2 = doc.layers.new("Layer2")
    handles = [l1.dxf.handle, l2.dxf.handle]
    viewport = Viewport.from_text(ENTITY_R2000)
    # implant some handles
    viewport.frozen_layers = handles
    result = viewport.post_load_hook(doc)
    assert result is None
    assert viewport.frozen_layers == [
        "Layer1",
        "Layer2",
    ], "Layer handles must be resolved"


class TestFrozenLayers:
    @pytest.fixture
    def vp(self):
        return Viewport.new("F000")

    def test_set_and_get__frozen_layer_names(self, vp):
        layer_names = ["bricks", "steel", "glass"]
        vp.frozen_layers = layer_names
        assert layer_names == vp.frozen_layers

    def test_freeze_a_specific_layer(self, vp):
        vp.freeze("MyLayer")
        assert vp.frozen_layers == ["MyLayer"]

    def test_freezing_a_layer_twice_does_not_duplicate_entry(self, vp):
        vp.freeze("MyLayer")
        vp.freeze("MYLAYER")
        assert vp.frozen_layers == ["MyLayer"]

    def test_is_a_specific_layer_frozen(self, vp):
        vp.freeze("MyLayer")
        assert vp.is_frozen("MyLayer") is True
        assert (
            vp.is_frozen("MYLAYER") is True
        ), "layer names are case insensitive!"
        assert vp.is_frozen("OTHER_LAYER") is False

    def test_thaw_a_specific_layer(self, vp):
        vp.freeze("MyLayer")
        vp.thaw("MYLAYER")
        assert len(vp.frozen_layers) == 0, "layer names are case insensitive!"

    def test_ignore_thawing_non_frozen_layers_silently(self, vp):
        vp.freeze("MyLayer")
        vp.thaw("LayerA")
        vp.thaw("LayerB")
        assert vp.frozen_layers == ["MyLayer"]


def test_is_top_view():
    vp: Viewport = Viewport.from_text(ENTITY_R2000)
    assert vp.is_top_view is True


def test_get_scale():
    vp: Viewport = Viewport.from_text(ENTITY_R2000)
    assert vp.get_scale() == 1.0


def test_get_transformation_matrix():
    vp: Viewport = Viewport.from_text(ENTITY_R2000)
    m = vp.get_transformation_matrix()
    assert m.transform((1, 2, 3)).isclose((1, 2, 3))


def test_has_clipping_path():
    vp: Viewport = Viewport.from_text(ENTITY_R2000)
    assert vp.has_extended_clipping_path is False


def test_clipping_path_corners():
    vp: Viewport = Viewport.from_text(ENTITY_R2000)
    # the clipping rectangle as corner vertices
    assert len(vp.clipping_rect_corners()) == 4


def test_clipping_rect():
    vp: Viewport = Viewport.from_text(ENTITY_R2000)
    assert len(vp.clipping_rect()) == 2


def test_aspect_ratio():
    vp: Viewport = Viewport.new()
    vp.dxf.width = 2.0
    vp.dxf.height = 1.0
    assert vp.get_aspect_ratio() == 2.0


def test_invalid_aspect_ratio():
    vp: Viewport = Viewport.new()
    vp.dxf.width = 2.0
    vp.dxf.height = 0.0
    assert vp.get_aspect_ratio() == 0.0


def test_modelspace_limits():
    vp: Viewport = Viewport.new(dxfattribs={
        "width": 3.0,
        "height": 2.0,
        "view_center_point": (5, 7),
        "view_height": 6.0,
    })
    # view_width = 4.0
    x0, y0, x1, y1 = vp.get_modelspace_limits()
    assert x0 == pytest.approx(0.5)
    assert y0 == pytest.approx(4.0)
    assert x1 == pytest.approx(9.5)
    assert y1 == pytest.approx(10.0)


def test_modelspace_limits_rotated():
    vp: Viewport = Viewport.new(dxfattribs={
        "width": 2.0,
        "height": 1.0,
        "view_center_point": (2, 2),
        "view_height": 2.0,
        "view_twist_angle": 90
    })
    # view_width = 4.0
    x0, y0, x1, y1 = vp.get_modelspace_limits()
    assert x0 == pytest.approx(1.0)
    assert y0 == pytest.approx(-4.0)
    assert x1 == pytest.approx(3.0)
    assert y1 == pytest.approx(0.0)



# Copyright (c) 2020-2022, Matthew Broadway
# License: MIT License
from typing import Callable
from typing_extensions import TypeAlias
from ezdxf.entities import DXFGraphic

LayerName: TypeAlias = str
Color: TypeAlias = str
Radians: TypeAlias = float
FilterFunc: TypeAlias = Callable[[DXFGraphic], bool]



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License
import pytest
from datetime import datetime
from ezdxf.acis import sat, hdr, const
from ezdxf.version import __version__


def test_default_header():
    header = hdr.AcisHeader()
    assert header.version == const.MIN_EXPORT_VERSION
    assert header.n_entities == 0
    assert header.n_records == 0
    assert isinstance(header.creation_date, datetime)


LEN = len(__version__) + 20

HEADER_400 = f"""400 0 1 0 
{LEN} ezdxf v{__version__} ACIS Builder 12 ACIS 4.00 NT 24 Sat Jan  1 10:00:00 2022 
1 9.9999999999999995e-007 1e-010 """

HEADER_700 = f"""700 0 1 0 
@{LEN} ezdxf v{__version__} ACIS Builder @12 ACIS 32.0 NT @24 Sat Jan  1 10:00:00 2022 
1 9.9999999999999995e-007 1e-010 """

HEADER_21800 = f"""21800 0 1 0 
@{LEN} ezdxf v{__version__} ACIS Builder @14 ACIS 218.00 NT @24 Sat Jan  1 10:00:00 2022 
1 9.9999999999999995e-007 1e-010 """


@pytest.mark.parametrize(
    "ver,s",
    [
        (400, HEADER_400),
        (700, HEADER_700),
        (21800, HEADER_21800),
    ],
)
def test_dump_header_string(ver, s):
    header = hdr.AcisHeader()
    header.set_version(ver)
    header.creation_date = datetime(2022, 1, 1, 10, 00)
    header.n_entities = 1
    data = header.dumps()

    assert "\n".join(header.dumps()) == s


@pytest.mark.parametrize(
    "s",
    [
        "18 ezdxf ACIS Builder 14 ACIS 208.00 NT 24 Sat Jan  1 10:00:00 2022 ",
        "@18 ezdxf ACIS Builder @14 ACIS 208.00 NT @24 Sat Jan  1 10:00:00 2022 ",
    ],
)
def test_parse_header_str(s):
    tokens = list(sat.parse_header_str(s))
    assert tokens == [
        "ezdxf ACIS Builder",
        "ACIS 208.00 NT",
        "Sat Jan  1 10:00:00 2022",
    ]


@pytest.mark.parametrize("hdr,ver", [(HEADER_400, 400), (HEADER_21800, 21800)])
def test_parse_sat_header(hdr, ver):
    header, data = sat.parse_header(hdr.split("\n"))
    assert len(data) == 0
    assert header.version == ver
    assert header.product_id == f"ezdxf v{__version__} ACIS Builder"
    assert header.n_entities == 1
    assert header.creation_date == datetime(2022, 1, 1, 10, 00)


class TestMergeRecordStrings:
    @pytest.mark.parametrize(
        "data",
        [
            ["test 1 2 3 #", "End-of-ACIS-data", "test 4 5 6"],
            ["test 4 5 6 #", "Begin-of-ACIS-History-data", "test 4 5 6"],
        ],
    )
    def test_end_of_records_detection(self, data):
        assert len(list(sat.merge_record_strings(data))) == 1

    @pytest.mark.parametrize(
        "data",
        [
            ["test 1 2 3 #", "test 1 2 3 #"],
            ["test 4 5 6 ", "1 2 3 #", "test 1 2 3 #"],
            ["test 4 5 6 #", "test 1 2 3 ", "4 5 6 #"],
        ],
    )
    def test_merge_records(self, data):
        assert len(list(sat.merge_record_strings(data))) == 2

    def test_weird_placement_of_record_terminator(self):
        """This example was found in a SAT exported by BricsCAD."""
        data = [
            "eye_refinement ... end_fields #",
            "integer_attrib-name_attrib-gen-attrib ... #torus-surface ...",
            "I I I I #",
        ]
        records = list(sat.merge_record_strings(data))
        assert records[0] == "eye_refinement ... end_fields"
        assert records[1] == "integer_attrib-name_attrib-gen-attrib ..."
        assert records[2] == "torus-surface ... I I I I"


class TestParseRecords:
    def test_simple_case(self):
        records = sat.parse_records(["test 1 2 3 #", "test 4 5 6 #"])
        assert len(records) == 2
        assert records[0] == ["test", "1", "2", "3"]
        assert records[1] == ["test", "4", "5", "6"]

    def test_sequence_numbers(self):
        records = sat.parse_records(["-0 test 1 2 3 #", " -1 test 4 5 6 #"])
        assert len(records) == 2
        assert records[0] == ["test", "1", "2", "3"]
        assert records[1] == ["test", "4", "5", "6"]

    def test_non_continuous_sequence_numbers_raises_exception(self):
        with pytest.raises(sat.ParsingError):
            sat.parse_records(["-0 test 1 2 3 #", " -2 test 4 5 6 #"])

    @pytest.mark.parametrize(
        "data",
        [
            ["sentinel 7 #", "test 1 2 3 4 5 6 #", "sentinel 8 #"],
            ["sentinel 7 #", "test 1 2 3", " 4 5 6 #", "sentinel 8 #"],
            [
                "sentinel 7 #",
                "test",
                "1",
                "2",
                "3",
                "4",
                "5",
                "6",
                "#",
                "sentinel 8 #",
            ],
        ],
    )
    def test_merged_records(self, data):
        records = sat.parse_records(data)
        assert records[0] == ["sentinel", "7"]
        assert records[1] == ["test", "1", "2", "3", "4", "5", "6"]
        assert records[2] == ["sentinel", "8"]


def test_build_entities(prism_sat):
    data = prism_sat.splitlines()
    records = sat.parse_records(data[3:])
    entities = sat.build_entities(records, 700)
    assert len(entities) == 113
    assert entities[0].name == "body"
    assert entities[112].name == "straight-curve"


class TestAcisBuilder:
    @pytest.fixture(scope="class")
    def builder(self, prism_sat):
        return sat.parse_sat(prism_sat)

    def test_parsing_result(self, builder):
        assert builder.header.version == 700
        assert len(builder.bodies) == 1
        assert len(builder.entities) == 113

    def test_body_entity(self, builder):
        body = builder.bodies[0]
        assert body.name == "body"
        assert len(body.data) == 4
        assert body.data[1].name == "lump", "string ptr should be resolved"

    def test_ptr_resolving(self, builder):
        assert builder.entities[0].name == "body"
        assert builder.entities[112].name == "straight-curve"

    def test_attr_ptr_is_reset(self, builder):
        assert all(e.attr_ptr == "$-1" for e in builder.entities)

    def test_dump_sat_recreates_the_source_structure(self, builder, prism_sat):
        lines = builder.dump_sat()
        assert len(lines) == 117
        assert lines == prism_sat.splitlines()


def test_build_str_records():
    a = sat.new_entity("test1", data=[sat.NULL_PTR, 1])
    b = sat.new_entity("test2", id=7, data=[a, 2.0])
    c = sat.new_entity("test3", data=[a, b])
    entities = [a, b, c]
    s = list(sat.build_str_records(entities, 700))
    assert s[0] == "test1 $-1 -1 $-1 1 #"
    assert s[1] == "test2 $-1 7 $0 2.0 #"
    assert s[2] == "test3 $-1 -1 $0 $1 #"


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021-2024, Manfred Moitzi
#  License: MIT License

import pytest
from itertools import product
import math
import numpy as np

from ezdxf.math import Matrix44, Vec3
from ezdxf.entities import Ellipse

UNIFORM_SCALING = [
    (2, 2, 2),
    (-1, 1, 1),
    (1, -1, 1),
    (1, 1, -1),
    (-2, -2, 2),
    (2, -2, -2),
    (-2, 2, -2),
    (-3, -3, -3),
]
NON_UNIFORM_SCALING = [
    (-1, 2, 3),
    (1, -2, 3),
    (1, 2, -3),
    (-3, -2, 1),
    (3, -2, -1),
    (-3, 2, -1),
    (-3, -2, -1),
]

# chose fixed "random" values to narrow error space:
DELTA = [-2, -1, 0, 1, 2]


def synced_transformation(entity, chk, m: Matrix44):
    entity = entity.copy()
    entity.transform(m)
    chk = list(m.transform_vertices(chk))
    return entity, chk


def synced_scaling(entity, chk, sx=1, sy=1, sz=1):
    entity = entity.copy()
    entity.scale(sx, sy, sz)
    chk = list(Matrix44.scale(sx, sy, sz).transform_vertices(chk))
    return entity, chk


def build(angle, dx, dy, dz, axis, start, end, count):
    ellipse = Ellipse.new(
        dxfattribs={
            "start_param": start,
            "end_param": end,
        }
    )
    vertices = list(ellipse.vertices(ellipse.params(count)))
    m = Matrix44.chain(
        Matrix44.axis_rotate(axis=axis, angle=angle),
        Matrix44.translate(dx=dx, dy=dy, dz=dz),
    )
    return synced_transformation(ellipse, vertices, m)


def check(ellipse, vertices, count):
    ellipse_vertices = list(ellipse.vertices(ellipse.params(count)))
    # Ellipse vertices may appear in reverse order
    if not vertices[0].isclose(ellipse_vertices[0], abs_tol=1e-5):
        ellipse_vertices.reverse()

    return all(
        vtx.isclose(chk, abs_tol=1e-5)
        for vtx, chk in zip(ellipse_vertices, vertices)
    )


@pytest.mark.parametrize("sx, sy, sz", UNIFORM_SCALING + NON_UNIFORM_SCALING)
@pytest.mark.parametrize(
    "start, end",
    [
        # closed ellipse fails at non uniform scaling test, because no start-
        # and end param adjustment is applied, so generated vertices do not
        # match test vertices.
        (0, math.pi),  # half ellipse as special case
        (math.pi / 6, math.pi / 6 * 11),  # start < end
        (math.pi / 6 * 11, math.pi / 6),  # start > end
    ],
)
def test_random_ellipse_transformations(sx, sy, sz, start, end):
    vertex_count = 8

    for angle in np.linspace(0, math.tau, 19):
        for dx, dy, dz in product([2, 0, -2], repeat=3):
            axis = Vec3.random()

            config = (
                f"CONFIG sx={sx}, sy={sy}, sz={sz}; "
                f"start={start:.4f}, end={end:.4f}; angle={angle};"
                f"dx={dx}, dy={dy}, dz={dz}; axis={str(axis)}"
            )
            ellipse0, vertices0 = build(
                angle, dx, dy, dz, axis, start, end, vertex_count
            )
            assert check(ellipse0, vertices0, vertex_count) is True, config
            ellipse1, vertices1 = synced_scaling(
                ellipse0, vertices0, sx, sy, sz
            )
            assert check(ellipse1, vertices1, vertex_count) is True, config


# Detected error conditions:
ERROR_CONFIGS = [
    dict(
        scale=(-3, 2, -1),
        start=5.7596,
        end=0.5236,
        shift=(2, 2, 2),
        axis=(0.9071626602531079, 0.30514034769743803, -0.28973311175906485),
        angles=[4.886921],
    )  # Error occurs only for angle = 280 deg!
]


@pytest.mark.parametrize("config", ERROR_CONFIGS)
def test_error_config(config):
    count = 8
    sx, sy, sz = config["scale"]
    dx, dy, dz = config["shift"]
    start = config["start"]
    end = config["end"]
    axis = config["axis"]
    for angle in config["angles"]:
        ellipse0, vertices0 = build(angle, dx, dy, dz, axis, start, end, count)
        assert check(ellipse0, vertices0, count) is True
        ellipse1, vertices1 = synced_scaling(ellipse0, vertices0, sx, sy, sz)
        assert check(ellipse1, vertices1, count) is True


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2011-2019, Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.entities.ucs import UCSTableEntry


@pytest.fixture
def ucs():
    return UCSTableEntry.new(
        "FFFF",
        dxfattribs={
            "name": "UCS+90",
            "origin": (1.0, 1.0, 1.0),
            "xaxis": (0.0, 1.0, 0.0),
            "yaxis": (-1.0, 0.0, 0.0),
        },
    )


def test_name(ucs):
    assert ucs.dxf.name == "UCS+90"


def test_origin(ucs):
    assert ucs.dxf.origin == (1.0, 1.0, 1.0)


def test_xaxis(ucs):
    assert ucs.dxf.xaxis == (0.0, 1.0, 0.0)


def test_yaxis(ucs):
    assert ucs.dxf.yaxis == (-1.0, 0.0, 0.0)


def test_ucs(ucs):
    coords = ucs.ucs()
    assert coords.origin == (1, 1, 1)
    assert coords.ux == (0, 1, 0)
    assert coords.uy == (-1, 0, 0)
    assert coords.is_cartesian is True


def test_default_vales():
    ucs = UCSTableEntry.new()
    assert ucs.dxf.origin == (0, 0, 0)
    assert ucs.dxf.xaxis == (1, 0, 0)
    assert ucs.dxf.yaxis == (0, 1, 0)



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License

import os
import pytest
import ezdxf
from ezdxf import recover

BASEDIR = os.path.dirname(__file__)
DATADIR = "data"


@pytest.fixture(params=["duplicate_handles.dxf"])
def filename(request):
    filename = os.path.join(BASEDIR, DATADIR, request.param)
    if not os.path.exists(filename):
        pytest.skip(f"File {filename} not found.")
    return filename


def test_recover_duplicate_handles(filename, tmp_path):
    doc, auditor = recover.readfile(filename)
    fixed_dxf_name = tmp_path / "fixed.dxf"
    doc.saveas(fixed_dxf_name)
    assert doc.layers.head.dxf.handle != "2"
    assert doc.linetypes.head.dxf.handle != "4"

    doc2 = ezdxf.readfile(fixed_dxf_name)
    auditor = doc2.audit()
    assert len(auditor.fixes) == 0
    assert len(doc2.modelspace()) == 132


if __name__ == '__main__':
    pytest.main([__file__])



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.lldxf.types import DXFTag
from ezdxf.entities.dxfobj import XRecord
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text

XRECORD = """  0
XRECORD
5
0
330
0
100
AcDbXrecord
280
1
"""


@pytest.fixture
def entity():
    return XRecord.from_text(XRECORD)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "XRECORD" in ENTITY_CLASSES


def test_default_init():
    entity = XRecord()
    assert entity.dxftype() == "XRECORD"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = XRecord.new(handle="ABBA", owner="0", dxfattribs={})
    assert entity.dxf.cloning == 1
    assert len(entity.tags) == 0


def test_load_from_text(entity):
    assert entity.dxf.cloning == 1
    assert len(entity.tags) == 0


def test_write_dxf():
    entity = XRecord.from_text(XRECORD)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(XRECORD)
    assert result == expected


@pytest.fixture
def xrecord():
    return XRecord.from_text(XRECORD1)


def test_handle(xrecord):
    assert "43A" == xrecord.dxf.handle


def test_parent_handle(xrecord):
    assert "28C" == xrecord.dxf.owner


def test_cloning_parameter(xrecord):
    assert 1 == xrecord.dxf.cloning


def test_get_data(xrecord):
    assert DXFTag(102, "SHADEPLOT") == xrecord.tags[0]
    assert DXFTag(70, 0) == xrecord.tags[1]


def test_last_data(xrecord):
    assert DXFTag(70, 0) == xrecord.tags[-1]


def test_iter_data(xrecord):
    tags = list(xrecord.tags)
    assert DXFTag(102, "SHADEPLOT") == tags[0]
    assert DXFTag(70, 0) == tags[1]


def test_len(xrecord):
    assert 2 == len(xrecord.tags)


def test_set_data(xrecord):
    xrecord.tags[0] = DXFTag(103, "MOZMAN")
    assert DXFTag(103, "MOZMAN") == xrecord.tags[0]
    assert DXFTag(70, 0) == xrecord.tags[1]


def test_append_data(xrecord):
    xrecord.tags.append(DXFTag(103, "MOZMAN"))
    assert DXFTag(103, "MOZMAN") == xrecord.tags[-1]


def test_extend(xrecord):
    xrecord.extend([(103, "MOZMAN")])
    tag = xrecord.tags[-1]
    assert isinstance(tag, DXFTag) is True
    assert tag == DXFTag(103, "MOZMAN")


def test_clear(xrecord):
    xrecord.clear()
    assert len(xrecord.tags) == 0


def test_reset(xrecord):
    xrecord.reset([(1, "text"), (2, "text")])
    assert len(xrecord.tags) == 2
    assert xrecord.tags == [(1, "text"), (2, "text")]


XRECORD1 = """  0
XRECORD
  5
43A
102
{ACAD_REACTORS
330
28C
102
}
330
28C
100
AcDbXrecord
280
     1
102
SHADEPLOT
 70
     0
"""



# cython: language_level=3
# Copyright (c) 2020-2024, Manfred Moitzi
# License: MIT License

cdef bint isclose(double a, double b, double rel_tol, double abs_tol)
cdef double normalize_rad_angle(double a)
cdef double normalize_deg_angle(double a)

cdef class Vec2:
    cdef readonly double x, y

# Vec2 C-functions:
cdef Vec2 v2_add(Vec2 a, Vec2 b)
cdef Vec2 v2_sub(Vec2 a, Vec2 b)
cdef Vec2 v2_mul(Vec2 a, double factor)
cdef Vec2 v2_normalize(Vec2 a, double length)
cdef double v2_dot(Vec2 a, Vec2 b)
cdef double v2_det(Vec2 a, Vec2 b)
cdef double v2_dist(Vec2 a, Vec2 b)
cdef Vec2 v2_from_angle(double angle, double length)
cdef double v2_angle_between(Vec2 a, Vec2 b) except -1000
cdef Vec2 v2_lerp(Vec2 a, Vec2 b, double factor)
cdef Vec2 v2_ortho(Vec2 a, bint ccw)
cdef Vec2 v2_project(Vec2 a, Vec2 b)
cdef bint v2_isclose(Vec2 a, Vec2 b, double rel_tol, double abs_tol)


cdef class Vec3:
    cdef readonly double x, y, z

# Vec3 C-functions:
cdef Vec3 v3_add(Vec3 a, Vec3 b)
cdef Vec3 v3_sub(Vec3 a, Vec3 b)
cdef Vec3 v3_mul(Vec3 a, double factor)
cdef Vec3 v3_reverse(Vec3 a)
cdef double v3_dot(Vec3 a, Vec3 b)
cdef Vec3 v3_cross(Vec3 a, Vec3 b)
cdef double v3_magnitude_sqr(Vec3 a)
cdef double v3_magnitude(Vec3 a)
cdef double v3_dist(Vec3 a, Vec3 b)
cdef Vec3 v3_from_angle(double angle, double length)
cdef double v3_angle_between(Vec3 a, Vec3 b) except -1000
cdef double v3_angle_about(Vec3 a, Vec3 base, Vec3 target)
cdef Vec3 v3_normalize(Vec3 a, double length)
cdef Vec3 v3_lerp(Vec3 a, Vec3 b, double factor)
cdef Vec3 v3_ortho(Vec3 a, bint ccw)
cdef Vec3 v3_project(Vec3 a, Vec3 b)
cdef bint v3_isclose(Vec3 a, Vec3 b, double rel_tol, double abs_tol)




# Copyright (c) 2020-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Tuple, Iterable, Sequence
from typing_extensions import TypeAlias
import math
from ezdxf.math import Vec3, UVec

LineSegment: TypeAlias = Tuple[Vec3, Vec3]


class _LineTypeRenderer:
    """Renders a line segment as multiple short segments according to the given
    line pattern.

    In contrast to the DXF line pattern is this pattern simplified and follow
    the rule line-gap-line-gap-... a line of length 0 is a point.
    The pattern should end with a gap (even count) and the dash length is in
    drawing units.

    Args:
          dashes: sequence of floats, line-gap-line-gap-...

    """
    # Get the simplified line pattern by LineType.simplified_line_pattern()
    def __init__(self, dashes: Sequence[float]):
        self._dashes = dashes
        self._dash_count: int = len(dashes)
        self.is_solid: bool = True
        self._current_dash: int = 0
        self._current_dash_length: float = 0.0
        if self._dash_count > 1:
            self.is_solid = False
            self._current_dash_length = self._dashes[0]
            self._is_dash = True

    def line_segment(self, start: UVec, end: UVec) -> Iterable[LineSegment]:
        """Yields the line from `start` to `end` according to stored line
        pattern as short segments. Yields only the lines and points not the
        gaps.

        """
        _start = Vec3(start)
        _end = Vec3(end)
        if self.is_solid or _start.isclose(_end):
            yield _start, _end
            return

        segment_vec = _end - _start
        segment_length = segment_vec.magnitude
        segment_dir = segment_vec / segment_length  # normalize

        for is_dash, dash_length in self._render_dashes(segment_length):
            _end = _start + segment_dir * dash_length
            if is_dash:
                yield _start, _end
            _start = _end

    def _render_dashes(self, length: float) -> Iterable[tuple[bool, float]]:
        if length <= self._current_dash_length:
            self._current_dash_length -= length
            yield self._is_dash, length
            if math.isclose(self._current_dash_length, 0.0):
                self._cycle_dashes()
        else:
            # Avoid deep recursions!
            while length > self._current_dash_length:
                length -= self._current_dash_length
                yield from self._render_dashes(self._current_dash_length)
            if length > 0.0:
                yield from self._render_dashes(length)

    def _cycle_dashes(self):
        self._current_dash = (self._current_dash + 1) % self._dash_count
        self._current_dash_length = self._dashes[self._current_dash]
        self._is_dash = not self._is_dash



# Copyright (c) 2011-2018, Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.entities.view import View


@pytest.fixture
def view():
    return View.new(
        "FFFF",
        dxfattribs={
            "name": "VIEW1",
            "flags": 0,
            "height": 1.0,
            "width": 1.0,
            "center": (0, 0),
            "direction": (0, 0, -1),
            "target": (0, 0, 0),
            "focal_length": 1.0,
            "front_clipping": 0.0,
            "back_clipping": 0.0,
            "view_twist": 0.0,
            "view_mode": 0,
        },
    )


def test_view_attribute_access(view):
    assert view.dxf.name == "VIEW1"
    assert view.dxf.flags == 0
    assert view.dxf.height == 1.0
    assert view.dxf.width == 1.0
    assert view.dxf.center == (0, 0)
    assert view.dxf.direction == (0, 0, -1)
    assert view.dxf.target == (0, 0, 0)
    assert view.dxf.focal_length == 1.0
    assert view.dxf.front_clipping == 0
    assert view.dxf.back_clipping == 0
    assert view.dxf.view_twist == 0
    assert view.dxf.view_mode == 0



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
import pytest

from pathlib import Path
import ezdxf
from ezdxf.protocols import query_virtual_entities

DATA = Path(__file__).parent / "data"
ACAD_TABLE_WITH_BLK_REF = "acad_table_with_blk_ref.dxf"


@pytest.fixture(scope="module")
def doc():
    return ezdxf.readfile(DATA / ACAD_TABLE_WITH_BLK_REF)


@pytest.fixture
def acad_table(doc):
    msp = doc.modelspace()
    return msp.query("ACAD_TABLE").first


def test_virtual_entities_from_block(acad_table):
    content = query_virtual_entities(acad_table)
    assert len(content) == 42
    # Provides real INSERT entities:
    assert sum(int(e.dxftype() == "INSERT") for e in content) == 1


def test_virtual_entities_from_proxy_graphic(acad_table):
    content = list(acad_table.proxy_graphic_content())
    assert len(content) == 45
    # Does not provide INSERT entities:
    assert sum(int(e.dxftype() == "INSERT") for e in content) == 0


if __name__ == '__main__':
    pytest.main([__file__])



# Copyright (C) 2011-2019, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf

new = ezdxf.new


def test_new_AC1009():
    doc = new("R12")
    assert "AC1009" == doc.dxfversion


def test_new_AC1015():
    doc = new("R2000")
    assert "AC1015" == doc.dxfversion


def test_new_AC1018():
    doc = new("R2004")
    assert "AC1018" == doc.dxfversion


def test_new_AC1021():
    doc = new("R2007")
    assert "AC1021" == doc.dxfversion


def test_new_AC1024():
    doc = new("R2010")
    assert "AC1024" == doc.dxfversion


def test_new_AC1027():
    doc = new("R2013")
    assert "AC1027" == doc.dxfversion


def test_new_AC1032():
    doc = new("R2018")
    assert "AC1032" == doc.dxfversion


def test_invalid_dxf_version():
    with pytest.raises(ezdxf.const.DXFVersionError):
        new("R13")
    with pytest.raises(ezdxf.const.DXFVersionError):
        new("R14")
    with pytest.raises(ezdxf.const.DXFVersionError):
        new("XYZ")
    with pytest.raises(ezdxf.const.DXFVersionError):
        new("AC1012")
    with pytest.raises(ezdxf.const.DXFVersionError):
        new("AC1013")
    with pytest.raises(ezdxf.const.DXFVersionError):
        new("AC1014")



#  Copyright (c) 2022-2024, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import (
    Any,
    Sequence,
    Iterator,
    Union,
    List,
    TYPE_CHECKING,
    Optional,
)
from typing_extensions import TypeAlias
import math
from datetime import datetime

from . import const
from .const import ParsingError, InvalidLinkStructure
from .hdr import AcisHeader
from .abstract import (
    AbstractEntity,
    AbstractBuilder,
    DataLoader,
    DataExporter,
    EntityExporter,
)

if TYPE_CHECKING:
    from .entities import AcisEntity
    from ezdxf.math import Vec3

SatRecord: TypeAlias = List[str]


class SatEntity(AbstractEntity):
    """Low level representation of an ACIS entity (node)."""

    def __init__(
        self,
        name: str,
        attr_ptr: str = "$-1",
        id: int = -1,
        data: Optional[list[Any]] = None,
    ):
        self.name = name
        self.attr_ptr = attr_ptr
        self.id = id
        self.data: list[Any] = data if data is not None else []
        self.attributes: "SatEntity" = None  # type: ignore

    def __str__(self):
        return f"{self.name}({self.id})"


NULL_PTR = SatEntity("null-ptr", "$-1", -1, tuple())  # type: ignore


def new_entity(
    name: str,
    attributes=NULL_PTR,
    id=-1,
    data: Optional[list[Any]] = None,
) -> SatEntity:
    """Factory to create new ACIS entities.

    Args:
        name: entity type
        attributes: reference to the entity attributes or :attr:`NULL_PTR`.
        id: unique entity id as integer or -1
        data: generic data container as list

    """
    e = SatEntity(name, "$-1", id, data)
    e.attributes = attributes
    return e


def is_ptr(s: str) -> bool:
    """Returns ``True`` if the string `s` represents an entity pointer."""
    return len(s) > 0 and s[0] == "$"


def resolve_str_pointers(entities: list[SatEntity]) -> list[SatEntity]:
    def ptr(s: str) -> SatEntity:
        num = int(s[1:])
        if num == -1:
            return NULL_PTR
        return entities[num]

    for entity in entities:
        entity.attributes = ptr(entity.attr_ptr)
        entity.attr_ptr = "$-1"
        data = []
        for token in entity.data:
            if is_ptr(token):
                data.append(ptr(token))
            else:
                data.append(token)
        entity.data = data
    return entities


class SatDataLoader(DataLoader):
    def __init__(self, data: list[Any], version: int):
        self.version = version
        self.data = data
        self.index = 0

    def has_data(self) -> bool:
        return self.index <= len(self.data)

    def read_int(self, skip_sat: Optional[int] = None) -> int:
        if skip_sat is not None:
            return skip_sat

        entry = self.data[self.index]
        try:
            value = int(entry)
        except ValueError:
            raise ParsingError(f"expected integer, got {entry}")
        self.index += 1
        return value

    def read_double(self) -> float:
        entry = self.data[self.index]
        try:
            value = float(entry)
        except ValueError:
            raise ParsingError(f"expected double, got {entry}")
        self.index += 1
        return value

    def read_interval(self) -> float:
        finite = self.read_bool("F", "I")
        if finite:
            return self.read_double()
        return math.inf

    def read_vec3(self) -> tuple[float, float, float]:
        x = self.read_double()
        y = self.read_double()
        z = self.read_double()
        return x, y, z

    def read_bool(self, true: str, false: str) -> bool:
        value = self.data[self.index]
        if value == true:
            self.index += 1
            return True
        elif value == false:
            self.index += 1
            return False
        raise ParsingError(
            f"expected bool string '{true}' or '{false}', got {value}"
        )

    def read_str(self) -> str:
        value = self.data[self.index]
        if self.version < const.Features.AT or value.startswith("@"):
            self.index += 2
            return self.data[self.index - 1]
        raise ParsingError(f"expected string, got {value}")

    def read_ptr(self) -> AbstractEntity:
        entity = self.data[self.index]
        if isinstance(entity, AbstractEntity):
            self.index += 1
            return entity
        raise ParsingError(f"expected pointer, got {type(entity)}")

    def read_transform(self) -> list[float]:
        # 4th column is not stored
        # Read only the matrix values which contain all information needed,
        # the additional data are only hints for the kernel how to process
        # the data (rotation, reflection, scaling, shearing).
        return [self.read_double() for _ in range(12)]


class SatBuilder(AbstractBuilder):
    """Low level data structure to manage ACIS SAT data files."""

    def __init__(self) -> None:
        self.header = AcisHeader()
        self.bodies: list[SatEntity] = []
        self.entities: list[SatEntity] = []
        self._export_mapping: dict[int, SatEntity] = {}

    def dump_sat(self) -> list[str]:
        """Returns the text representation of the ACIS file as list of strings
        without line endings.

        Raise:
            InvalidLinkStructure: referenced ACIS entity is not stored in
                the :attr:`entities` storage

        """
        self.reorder_records()
        self.header.n_entities = len(self.bodies) + int(
            self.header.has_asm_header
        )
        if self.header.version == 700:
            self.header.n_records = 0  # ignored for old versions
        else:
            self.header.n_records = len(self.entities)
        data = self.header.dumps()
        data.extend(build_str_records(self.entities, self.header.version))
        data.append(self.header.sat_end_marker())
        return data

    def set_entities(self, entities: list[SatEntity]) -> None:
        """Reset entities and bodies list. (internal API)"""
        self.bodies = [e for e in entities if e.name == "body"]
        self.entities = entities


class SatExporter(EntityExporter[SatEntity]):
    def make_record(self, entity: AcisEntity) -> SatEntity:
        record = SatEntity(entity.type, id=entity.id)
        record.attributes = NULL_PTR
        return record

    def make_data_exporter(self, record: SatEntity) -> DataExporter:
        return SatDataExporter(self, record.data)

    def dump_sat(self) -> list[str]:
        builder = SatBuilder()
        builder.header = self.header
        builder.set_entities(self.export_records())
        return builder.dump_sat()


def build_str_records(entities: list[SatEntity], version: int) -> Iterator[str]:
    def ptr_str(e: SatEntity) -> str:
        if e is NULL_PTR:
            return "$-1"
        try:
            return f"${entities.index(e)}"
        except ValueError:
            raise InvalidLinkStructure(f"entity {str(e)} not in record storage")

    for entity in entities:
        tokens = [entity.name]
        tokens.append(ptr_str(entity.attributes))
        if version >= 700:
            tokens.append(str(entity.id))
        for data in entity.data:
            if isinstance(data, SatEntity):
                tokens.append(ptr_str(data))
            else:
                tokens.append(str(data))
        tokens.append("#")
        yield " ".join(tokens)


def parse_header_str(s: str) -> Iterator[str]:
    num = ""
    collect = 0
    token = ""
    for c in s.rstrip():
        if collect > 0:
            token += c
            collect -= 1
            if collect == 0:
                yield token
                token = ""
        elif c == "@":
            continue
        elif c in "0123456789":
            num += c
        elif c == " " and num:
            collect = int(num)
            num = ""


def parse_header(data: Sequence[str]) -> tuple[AcisHeader, Sequence[str]]:
    header = AcisHeader()
    tokens = data[0].split()
    header.version = int(tokens[0])
    try:
        header.n_records = int(tokens[1])
        header.n_entities = int(tokens[2])
        header.flags = int(tokens[3])
    except (IndexError, ValueError):
        pass
    tokens = list(parse_header_str(data[1]))
    try:
        header.product_id = tokens[0]
        header.acis_version = tokens[1]
    except IndexError:
        pass

    if len(tokens) > 2:
        try:  # Sat Jan  1 10:00:00 2022
            header.creation_date = datetime.strptime(tokens[2], const.DATE_FMT)
        except ValueError:
            pass
    tokens = data[2].split()
    try:
        header.units_in_mm = float(tokens[0])
    except (IndexError, ValueError):
        pass
    return header, data[3:]


def _filter_records(data: Sequence[str]) -> Iterator[str]:
    for line in data:
        if line.startswith(const.END_OF_ACIS_DATA_SAT) or line.startswith(
            const.BEGIN_OF_ACIS_HISTORY_DATA
        ):
            return
        yield line


def merge_record_strings(data: Sequence[str]) -> Iterator[str]:
    merged_data = " ".join(_filter_records(data))
    for record in merged_data.split("#"):
        record = record.strip()
        if record:
            yield record


def parse_records(data: Sequence[str]) -> list[SatRecord]:
    expected_seq_num = 0
    records: list[SatRecord] = []
    for line in merge_record_strings(data):
        tokens: SatRecord = line.split()
        first_token = tokens[0].strip()
        if first_token.startswith("-"):
            num = -int(first_token)
            if num != expected_seq_num:
                raise ParsingError(
                    "non-continuous sequence numbers not supported"
                )
            tokens.pop(0)
        records.append(tokens)
        expected_seq_num += 1
    return records


def build_entities(
    records: Sequence[SatRecord], version: int
) -> list[SatEntity]:
    entities: list[SatEntity] = []
    for record in records:
        name = record[0]
        attr = record[1]
        id_ = -1
        if version >= 700:
            id_ = int(record[2])
            data = record[3:]
        else:
            data = record[2:]
        entities.append(SatEntity(name, attr, id_, data))
    return entities


def parse_sat(s: Union[str, Sequence[str]]) -> SatBuilder:
    """Returns the :class:`SatBuilder` for the ACIS :term:`SAT` file content
    given as string or list of strings.

    Raises:
        ParsingError: invalid or unsupported ACIS data structure

    """
    data: Sequence[str]
    if isinstance(s, str):
        data = s.splitlines()
    else:
        data = s
    if not isinstance(data, Sequence):
        raise TypeError("expected as string or a sequence of strings")
    builder = SatBuilder()
    header, data = parse_header(data)
    builder.header = header
    records = parse_records(data)
    entities = build_entities(records, header.version)
    builder.set_entities(resolve_str_pointers(entities))
    return builder


class SatDataExporter(DataExporter):
    def __init__(self, exporter: SatExporter, data: list[Any]):
        self.version = exporter.version
        self.exporter = exporter
        self.data = data

    def write_int(self, value: int, skip_sat=False) -> None:
        """There are sometimes additional int values in SAB files which are
        not present in SAT files, maybe reference counters e.g. vertex, coedge.
        """
        if not skip_sat:
            self.data.append(str(value))

    def write_double(self, value: float) -> None:
        self.data.append(f"{value:g}")

    def write_interval(self, value: float) -> None:
        if math.isinf(value):
            self.data.append("I")  # infinite
        else:
            self.data.append("F")  # finite
            self.write_double(value)

    def write_loc_vec3(self, value: Vec3) -> None:
        self.write_double(value.x)
        self.write_double(value.y)
        self.write_double(value.z)

    def write_dir_vec3(self, value: Vec3) -> None:
        self.write_double(value.x)
        self.write_double(value.y)
        self.write_double(value.z)

    def write_bool(self, value: bool, true: str, false: str) -> None:
        self.data.append(true if value else false)

    def write_str(self, value: str) -> None:
        self.data.append(f"@{len(value)}")
        self.data.append(str(value))

    def write_literal_str(self, value: str) -> None:
        self.write_str(value)  # just for SAB files important

    def write_ptr(self, entity: AcisEntity) -> None:
        record = NULL_PTR
        if not entity.is_none:
            record = self.exporter.get_record(entity)
        self.data.append(record)

    def write_transform(self, data: list[str]) -> None:
        self.data.extend(data)



# Copyright (C) 2011-2022, Manfred Moitzi
# License: MIT License
from .mtextsurrogate import MTextSurrogate
from .tablepainter import TablePainter, CustomCell
from .menger_sponge import MengerSponge
from .sierpinski_pyramid import SierpinskyPyramid
from .dimlines import LinearDimension, AngularDimension, ArcDimension, RadialDimension, dimstyles
from .importer import Importer
from .r12writer import r12writer
from .mtxpl import MTextExplode



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License

import pytest

from ezdxf.npshapes import NumpyPoints2d, NumpyPath2d, NumpyPoints3d
from ezdxf.math import Matrix44, BoundingBox2d, close_vectors, Vec2, Vec3
from ezdxf.path import Command, from_vertices, Path
from ezdxf.render import forms


class TestNumpyPoints2d:
    @pytest.fixture
    def points(self):
        return Vec2.list([(1, 2), (7, 4), (4, 7), (0, 1)])

    def test_conversion(self, points):
        pl = NumpyPoints2d(points)
        assert len(pl) == len(points)
        assert all(v0.isclose(v1) for v0, v1 in zip(pl.vertices(), points))

    def test_extents(self, points):
        pl = NumpyPoints2d(points)
        extmin, extmax = pl.extents()
        assert extmin.isclose((0, 1))
        assert extmax.isclose((7, 7))

    def test_transform_inplace(self, points):
        m = Matrix44.translate(7, 8, 0)
        t_pts = m.fast_2d_transform(points)
        pl = NumpyPoints2d(points)
        pl.transform_inplace(m)
        assert all(v0.isclose(v1) for v0, v1 in zip(pl.vertices(), t_pts))

    def test_to_tuples(self, points):
        pl = NumpyPoints2d(points)
        vertices = pl.to_tuples()
        assert isinstance(vertices, list)
        assert len(vertices) == len(points)
        assert isinstance(vertices[0], tuple)
        assert len(vertices[0]) == 2

    def test_to_list(self, points):
        pl = NumpyPoints2d(points)
        vertices = pl.to_list()
        assert isinstance(vertices, list)
        assert len(vertices) == len(points)
        assert isinstance(vertices[0], list)
        assert len(vertices[0]) == 2


class TestNumpyPoints3d:
    @pytest.fixture
    def points(self):
        return Vec3.list([(1, 2, 3), (7, 4, 3), (4, 7, 9), (0, 1, 3)])

    def test_conversion(self, points):
        pl = NumpyPoints3d(points)
        assert len(pl) == len(points)
        assert all(v0.isclose(v1) for v0, v1 in zip(pl.vertices(), points))

    def test_extents(self, points):
        pl = NumpyPoints3d(points)
        extmin, extmax = pl.extents()
        assert extmin.isclose((0, 1, 3))
        assert extmax.isclose((7, 7, 9))

    def test_transform_inplace(self, points):
        m = Matrix44.translate(7, 8, 0)
        t_pts = list(m.transform_vertices(points))
        pl = NumpyPoints3d(points)
        pl.transform_inplace(m)
        assert all(v0.isclose(v1) for v0, v1 in zip(pl.vertices(), t_pts))


class TestNumpyPath2d:
    @pytest.fixture
    def path(self):
        p = Path((1, 2))
        p.line_to((7, 4))
        p.curve3_to((4, 7), (0, 1))
        p.move_to((10, 0))
        p.curve4_to((15, 7), (13, 3), (14, 5))
        return p

    def test_clone(self, path):
        np_path = NumpyPath2d(path)
        clone_ = np_path.clone().to_path()
        assert clone_.control_vertices() == path.control_vertices()
        assert clone_.command_codes() == path.command_codes()

    def test_start_point(self, path):
        assert path.start.isclose((1, 2))

    def test_end_point(self, path):
        assert path.end.isclose((15, 7))

    def test_has_subpaths(self, path):
        np_path = NumpyPath2d(path)
        assert np_path.has_sub_paths is True

    def test_has_no_subpaths(self):
        np_path = NumpyPath2d(Path((1, 2)))
        assert np_path.has_sub_paths is False

    def test_to_path_2d(self, path):
        np_path = NumpyPath2d(path)
        assert len(np_path) == len(path)

        path = np_path.to_path()
        assert len(path) == 4
        assert path.start.isclose((1, 2))
        assert path.end.isclose((15, 7))

        cmds = path.commands()
        assert cmds[0].type == Command.LINE_TO
        assert cmds[0].end.isclose((7, 4))

        assert cmds[1].type == Command.CURVE3_TO
        assert cmds[1].end.isclose((4, 7))
        assert cmds[1].ctrl.isclose((0, 1))

        assert cmds[2].type == Command.MOVE_TO
        assert cmds[2].end.isclose((10, 0))

        assert cmds[3].type == Command.CURVE4_TO
        assert cmds[3].end.isclose((15, 7))
        assert cmds[3].ctrl1.isclose((13, 3))
        assert cmds[3].ctrl2.isclose((14, 5))

    def test_extents(self, path):
        np_path = NumpyPath2d(path)
        extmin, extmax = np_path.extents()
        box = BoundingBox2d(path.control_vertices())

        assert extmin.isclose(box.extmin)
        assert extmax.isclose(box.extmax)

    def test_transform(self, path):
        m = Matrix44.scale(2, 3, 1) @ Matrix44.translate(-2, 10, 0)
        np_path = NumpyPath2d(path)
        np_path.transform_inplace(m)
        assert all(
            v0.isclose(v1)
            for v0, v1 in zip(np_path.vertices(), path.transform(m).control_vertices())
        )

    def test_start_point_only_path(self):
        p = NumpyPath2d(Path((10, 20)))
        assert p.start.isclose((10, 20))
        # and back
        assert p.to_path().start.isclose((10, 20))

    def test_from_empty_path(self):
        p = NumpyPath2d(Path())
        assert len(p) == 0
        assert p.start == (0, 0)
        assert p.end == (0, 0)
        # and back
        assert p.to_path().start == (0, 0)  # default start point

    def test_create_empty_path_from_none(self):
        p = NumpyPath2d(None)
        assert len(p) == 0
        with pytest.raises(IndexError):
            assert p.start
        with pytest.raises(IndexError):
            assert p.end
        # and back
        assert p.to_path().start == (0, 0)  # default start point


@pytest.fixture
def first():
    p = Path()
    p.line_to((10, 0))
    return NumpyPath2d(p)


@pytest.fixture
def second():
    p = Path((10, 0))
    p.line_to((20, 0))
    return NumpyPath2d(p)


@pytest.fixture
def third():
    p = Path((20, 0))
    p.line_to((30, 0))
    return NumpyPath2d(p)


@pytest.fixture
def curve3():
    p = Path((0, 0))
    p.curve3_to((10, 0), (5, 3))
    return NumpyPath2d(p)


@pytest.fixture
def curve4():
    p = Path((10, 0))
    p.curve4_to((20, 0), (13, -3), (17, 3))
    return NumpyPath2d(p)


class TestNumpyPath2dExtend:
    def test_extend_empty_path(self, second):
        empty = NumpyPath2d(None)
        empty.extend([second])
        assert len(empty) == 1
        assert empty.start.isclose((10, 0))
        assert empty.end.isclose((20, 0))

    def test_extend_by_empty_path(self, first):
        first.extend([NumpyPath2d(None)])
        assert len(first) == 1
        assert first.start.isclose((0, 0))
        assert first.end.isclose((10, 0))

    def test_extend_by_empty_2d_path(self, first):
        empty = Path(Vec2(7, 7))  # has no drawing commands
        first.extend([NumpyPath2d(empty)])
        assert len(first) == 1
        assert first.start.isclose((0, 0))
        assert first.end.isclose((10, 0))

    def test_extend_by_empty_list(self, first):
        first.extend([])
        assert len(first) == 1
        assert first.start.isclose((0, 0))
        assert first.end.isclose((10, 0))

    def test_concatenate_adjacent_paths(self, first, second):
        base = NumpyPath2d.concatenate([first, second])
        assert base.command_codes() == [1, 1]
        assert base.start.isclose((0, 0))
        assert base.end.isclose((20, 0))

    def test_concatenate_separated_paths(self, first, third):
        base = NumpyPath2d.concatenate([first, third])
        assert base.has_sub_paths is True, "expected a MOVE_TO command"
        assert base.command_codes() == [1, 4, 1]
        vertices = base.vertices()
        assert len(vertices) == 4
        assert vertices[0].isclose((0, 0))  # start
        assert vertices[1].isclose((10, 0))  # line_to
        assert vertices[2].isclose((20, 0))  # move_to
        assert vertices[3].isclose((30, 0))  # line_to

    def test_concatenate_all_paths(self, first, second, third):
        base = NumpyPath2d.concatenate([first, second, third])
        assert base.command_codes() == [1, 1, 1]
        vertices = base.vertices()
        assert len(vertices) == 4
        assert vertices[0].isclose((0, 0))  # start
        assert vertices[1].isclose((10, 0))  # line_to
        assert vertices[2].isclose((20, 0))  # line_to
        assert vertices[3].isclose((30, 0))  # line_to

    def test_concatenate_curves(self, curve3, curve4):
        base = NumpyPath2d.concatenate([curve3, curve4])
        assert base.command_codes() == [2, 3]
        vertices = base.vertices()
        assert len(vertices) == 6
        assert vertices[0].isclose((0, 0))  # start
        assert vertices[1].isclose((5, 3))  # curve_3_to - ctrl
        assert vertices[2].isclose((10, 0))  # curve_3_to - end
        assert vertices[3].isclose((13, -3))  # curve_4_to - ctrl1
        assert vertices[4].isclose((17, 3))  # curve_4_to - ctrl2
        assert vertices[5].isclose((20, 0))  # curve_4_to - end

    def test_concatenate_empty_list_returns_empty_path(self):
        base = NumpyPath2d.concatenate([])
        assert base.command_codes() == []
        assert base.vertices() == []
        with pytest.raises(IndexError):
            base.start.isclose((0, 0))
        with pytest.raises(IndexError):
            base.end.isclose((0, 0))


class TestSubPaths:
    def test_empty_path(self):
        paths = NumpyPath2d(None).sub_paths()
        assert len(paths) == 0

    def test_single_path(self, first):
        paths = first.sub_paths()
        assert len(paths) == 1
        assert paths[0] is first

    def test_multipath_of_two(self, first, third):
        multi_path = NumpyPath2d.concatenate([first, third])
        paths = multi_path.sub_paths()
        assert len(paths) == 2
        for p in paths:
            assert p.command_codes() == [1]

        first, second = paths
        vertices = first.vertices()
        assert len(vertices) == 2
        assert vertices[0].isclose((0, 0))
        assert vertices[1].isclose((10, 0))

        vertices = second.vertices()
        assert len(vertices) == 2
        assert vertices[0].isclose((20, 0))
        assert vertices[1].isclose((30, 0))

    def test_multipath_with_curve3(self, first, curve3, third):
        multi_path = NumpyPath2d.concatenate([first, curve3, third])
        paths = multi_path.sub_paths()
        assert len(paths) == 3
        first, second, third = paths
        assert first.command_codes() == [1]
        assert second.command_codes() == [2]
        assert third.command_codes() == [1]

        vertices = first.vertices()
        assert len(vertices) == 2
        assert vertices[0].isclose((0, 0))
        assert vertices[1].isclose((10, 0))

        vertices = second.vertices()
        assert len(vertices) == 3
        assert vertices[0].isclose((0, 0))  # curve3_to, start
        assert vertices[1].isclose((5, 3))  # curve3_to, ctrl
        assert vertices[2].isclose((10, 0))  # curve3_to, end

        vertices = third.vertices()
        assert len(vertices) == 2
        assert vertices[0].isclose((20, 0))
        assert vertices[1].isclose((30, 0))

    def test_multipath_with_curve4(self, curve4, third, first):
        # curve4 and third are connected as a single path
        multi_path = NumpyPath2d.concatenate([curve4, third, first])
        paths = multi_path.sub_paths()
        assert len(paths) == 2
        first, second = paths
        assert first.command_codes() == [3, 1]
        assert second.command_codes() == [1]

        vertices = first.vertices()  # curve3 + third
        assert len(vertices) == 5
        assert vertices[0].isclose((10, 0))  # curve4_to, start
        assert vertices[1].isclose((13, -3))  # curve4_to, ctrl1
        assert vertices[2].isclose((17, 3))  # curve4_to, ctrl2
        assert vertices[3].isclose((20, 0))  # curve4_to, end
        assert vertices[4].isclose((30, 0))  # line_to

        vertices = second.vertices()
        assert len(vertices) == 2
        assert vertices[0].isclose((0, 0))
        assert vertices[1].isclose((10, 0))

    def test_sub_paths_are_reversible(self, first, third):
        multi_path = NumpyPath2d.concatenate([first, third])
        paths = multi_path.sub_paths()
        first, second = paths
        first.reverse()
        vertices = first.vertices()
        assert vertices[0].isclose((10, 0))
        assert vertices[1].isclose((0, 0))


def test_path_conversion_methods():
    source_path = from_vertices(forms.circle(32))
    p0 = Path((2, 0))
    p0.curve3_to((3, 0), (2.5, 1))
    source_path.extend_multi_path(p0)
    p0 = Path((3, 0))
    p0.curve4_to((4, 0), (3.3, -1), (3.7, 1))

    p0 = from_vertices(forms.translate(forms.circle(32)), (5, 0))
    source_path.extend_multi_path(p0)
    assert source_path.has_sub_paths is True
    assert source_path.has_curves is True
    assert source_path.has_lines is True

    converted_path = NumpyPath2d(source_path).to_path()
    assert converted_path.has_sub_paths is True
    assert converted_path.has_curves is True
    assert converted_path.has_lines is True

    assert source_path.start.isclose(converted_path.start)
    assert source_path.end.isclose(converted_path.end)

    assert source_path.command_codes() == converted_path.command_codes()
    cv0 = source_path.control_vertices()
    cv1 = converted_path.control_vertices()
    assert len(cv0) == len(cv1)
    for v0, v1 in zip(cv0, cv1):
        assert v0.isclose(v1)
    assert source_path._start_index == converted_path._start_index


@pytest.fixture(scope="module")
def p1():
    path = Path()
    path.line_to((2, 0))
    path.curve4_to((4, 0), (2, 1), (4, 1))  # end, ctrl1, ctrl2
    path.curve3_to((6, 0), (5, -1))  # end, ctrl
    return path


def test_flatten_path(p1):
    p2 = NumpyPath2d(p1)
    v1 = list(p1.flattening(0.01))
    v2 = list(p2.flattening(0.01))
    assert close_vectors(v1, v2)


class TestReversePath:
    def test_reversing_empty_path(self):
        p = NumpyPath2d(None)
        p.reverse()
        assert len(p) == 0

    def test_reversing_one_line(self):
        p = Path()
        p.line_to((1, 0))
        p2 = NumpyPath2d(p).reverse()
        vertices = p2.control_vertices()
        assert close_vectors(vertices, [(1, 0), (0, 0)])

    def test_reversing_one_curve3(self):
        p = Path()
        p.curve3_to((3, 0), (1.5, 1))
        p2 = NumpyPath2d(p).reverse()
        assert close_vectors(p2.control_vertices(), [(3, 0), (1.5, 1), (0, 0)])

    def test_reversing_one_curve4(self):
        p = Path()
        p.curve4_to((3, 0), (1, 1), (2, 1))
        p2 = NumpyPath2d(p).reverse()
        assert close_vectors(p2.control_vertices(), [(3, 0), (2, 1), (1, 1), (0, 0)])

    def test_reversing_path_ctrl_vertices(self, p1):
        p2 = NumpyPath2d(p1).reverse()
        assert close_vectors(
            p2.control_vertices(), reversed(list(p1.control_vertices()))
        )

    def test_reversing_flattened_path(self, p1):
        p2 = NumpyPath2d(p1)
        p2.reverse()
        v1 = list(p1.flattening(0.01))
        v2 = list(p2.flattening(0.01))
        assert close_vectors(v1, reversed(v2))

    def test_reversing_multi_path(self):
        p = Path()
        p.line_to((1, 0, 0))
        p.move_to((2, 0, 0))
        p.line_to((3, 0, 0))

        r = NumpyPath2d(p).reverse()
        assert r.has_sub_paths is True
        assert len(r) == 3
        assert r.command_codes() == [1, 4, 1]
        assert r.start == (3, 0, 0)
        assert r.end == (0, 0, 0)

    def test_reversing_multi_path_with_a_move_to_cmd_at_the_end(self):
        p = Path()
        p.line_to((1, 0, 0))
        p.move_to((2, 0, 0))
        # The last move_to will become the first move_to.
        # A move_to as first command just moves the start point.
        r = NumpyPath2d(p).reverse()
        assert len(r) == 1
        assert r.command_codes() == [1]
        assert r.start == (1, 0, 0)
        assert r.end == (0, 0, 0)
        assert r.has_sub_paths is False

    def test_has_clockwise_orientation(self, p1):
        p2 = NumpyPath2d(p1)
        assert p2.has_clockwise_orientation() is True

    def test_has_counter_clockwise_orientation(self, p1):
        p2 = NumpyPath2d(p1)
        assert p2.reverse().has_clockwise_orientation() is False

    def test_cw_and_ccw_orientation(self, p1):
        from ezdxf.math import has_clockwise_orientation

        p2 = NumpyPath2d(p1)
        assert has_clockwise_orientation(p2.clockwise().control_vertices()) is True
        assert (
            has_clockwise_orientation(p2.counter_clockwise().control_vertices())
            is False
        )


def test_clockwise_orientation_of_implicit_closed_path():
    p2 = NumpyPath2d(from_vertices([(0, 0), (10, 0), (10, 10), (0, 10)]))
    assert p2.has_clockwise_orientation() is False


def test_clockwise_orientation_of_explicit_closed_path():
    p2 = NumpyPath2d(from_vertices([(0, 0), (10, 0), (10, 10), (0, 10)], close=True))
    assert p2.has_clockwise_orientation() is False


def test_counter_clockwise_orientation_of_implicit_closed_path():
    p2 = NumpyPath2d(from_vertices([(0, 10), (10, 10), (10, 0), (0, 0)]))
    assert p2.has_clockwise_orientation() is True


def test_counter_clockwise_orientation_of_explicit_closed_path():
    p2 = NumpyPath2d(from_vertices([(0, 10), (10, 10), (10, 0), (0, 0)], close=True))
    assert p2.has_clockwise_orientation() is True


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2023, Manfred Moitzi
# License: MIT-License
from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from typing_extensions import Self
import copy
from ezdxf.lldxf import validator
from ezdxf.lldxf.const import SUBCLASS_MARKER, DXF2000, DXFStructureError
from ezdxf.lldxf.attributes import (
    DXFAttributes,
    DefSubclass,
    DXFAttr,
    group_code_mapping,
)
from ezdxf.lldxf.tags import Tags
from .dxfentity import base_class, SubclassProcessor
from .dxfobj import DXFObject
from .factory import register_entity
from .copy import default_copy

if TYPE_CHECKING:
    from ezdxf.entities import DXFNamespace, DXFEntity
    from ezdxf.lldxf.tagwriter import AbstractTagWriter

__all__ = ["VisualStyle"]

acdb_visualstyle = DefSubclass(
    "AcDbVisualStyle",
    {
        "description": DXFAttr(2),
        # Style type:
        # 0 = Flat
        # 1 = FlatWithEdges
        # 2 = Gouraud
        # 3 = GouraudWithEdges
        # 4 = 2dWireframe
        # 5 = Wireframe
        # 6 = Hidden
        # 7 = Basic
        # 8 = Realistic
        # 9 = Conceptual
        # 10 = Modeling
        # 11 = Dim
        # 12 = Brighten
        # 13 = Thicken
        # 14 = Linepattern
        # 15 = Facepattern
        # 16 = ColorChange
        # 20 = JitterOff
        # 21 = OverhangOff
        # 22 = EdgeColorOff
        # 23 = Shades of Gray
        # 24 = Sketchy
        # 25 = X-Ray
        # 26 = Shaded with edges
        # 27 = Shaded
        "style_type": DXFAttr(
            70, validator=validator.is_in_integer_range(0, 28)
        ),
        # Face lighting type:
        # 0 = Invisible
        # 1 = Visible
        # 2 = Phong
        # 3 = Gooch
        "face_lighting_model": DXFAttr(
            71, validator=validator.is_in_integer_range(0, 4)
        ),
        # Face lighting quality:
        # 0 = No lighting
        # 1 = Per face lighting
        # 2 = Per vertex lighting
        # 3 = Unknown
        "face_lighting_quality": DXFAttr(
            72, validator=validator.is_in_integer_range(0, 4)
        ),
        # Face color mode:
        # 0 = No color
        # 1 = Object color
        # 2 = Background color
        # 3 = Custom color
        # 4 = Mono color
        # 5 = Tinted
        # 6 = Desaturated
        "face_color_mode": DXFAttr(
            73, validator=validator.is_in_integer_range(0, 6)
        ),
        # Face modifiers:
        # 0 = No modifiers
        # 1 = Opacity
        # 2 = Specular
        # 3 = Unknown
        "face_modifiers": DXFAttr(
            90, validator=validator.is_in_integer_range(0, 4)
        ),
        "face_opacity_level": DXFAttr(40),
        "face_specular_level": DXFAttr(41),
        "color1": DXFAttr(62),
        "color2": DXFAttr(63),
        "face_style_mono_color": DXFAttr(421),
        # Edge style model:
        # 0 = No edges
        # 1 = Isolines
        # 2 = Facet edges
        "edge_style_model": DXFAttr(
            74, validator=validator.is_in_integer_range(0, 3)
        ),
        "edge_style": DXFAttr(91),
        "edge_intersection_color": DXFAttr(64),
        "edge_obscured_color": DXFAttr(65),
        "edge_obscured_linetype": DXFAttr(75),
        "edge_intersection_linetype": DXFAttr(175),
        "edge_crease_angle": DXFAttr(42),
        "edge_modifiers": DXFAttr(92),
        "edge_color": DXFAttr(66),
        "edge_opacity_level": DXFAttr(43),
        "edge_width": DXFAttr(76),
        "edge_overhang": DXFAttr(77),
        "edge_jitter": DXFAttr(78),
        "edge_silhouette_color": DXFAttr(67),
        "edge_silhouette_width": DXFAttr(79),
        "edge_halo_gap": DXFAttr(170),
        "edge_isoline_count": DXFAttr(171),
        "edge_hide_precision": DXFAttr(290),  # flag
        "edge_style_apply": DXFAttr(174),  # flag
        "style_display_settings": DXFAttr(93),
        "brightness": DXFAttr(44),
        "shadow_type": DXFAttr(173),
        "unknown1": DXFAttr(177),  # required if xdata is present?
        "internal_use_only_flag": DXFAttr(
            291
        ),  # visual style only use internal
        # Xdata must follow tag 291 (AutoCAD -> 'Xdata wasn't read' error)
        # 70: Xdata count (count of tag groups)
        # any code, value: multiple tags build a tag group
        # 176, 1: end of group marker
        # e.g. (291, 0) (70, 2) (62, 7) (420, 16777215) (176, 1) (90, 1) (176, 1)
    },
)
acdb_visualstyle_group_codes = group_code_mapping(acdb_visualstyle)


# undocumented Xdata in DXF R2018
@register_entity
class VisualStyle(DXFObject):
    """DXF VISUALSTYLE entity"""

    DXFTYPE = "VISUALSTYLE"
    DXFATTRIBS = DXFAttributes(base_class, acdb_visualstyle)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000  # official supported in R2007

    def __init__(self):
        super().__init__()
        self.acad_xdata = None  # to preserve AutoCAD xdata

    def copy_data(self, entity: Self, copy_strategy=default_copy) -> None:
        """Copy acad internal data."""
        assert isinstance(entity, VisualStyle)
        entity.acad_xdata = copy.deepcopy(self.acad_xdata)

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super().load_dxf_attribs(processor)
        if processor:
            tags = processor.subclass_by_index(1)
            if tags:
                self.acad_xdata = self.store_acad_xdata(tags)
                processor.fast_load_dxfattribs(
                    dxf, acdb_visualstyle_group_codes, subclass=tags
                )
            else:
                raise DXFStructureError(
                    f"missing 'AcDbVisualStyle' subclass in VISUALSTYLE(#{dxf.handle})"
                )

        return dxf

    @staticmethod
    def store_acad_xdata(tags: Tags):
        try:
            index = tags.tag_index(291)
        except IndexError:
            return None
        else:  # store tags after 291
            index += 1
            xdata = tags[index:]
            del tags[index:]  # remove xdata from subclass
            return xdata

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_visualstyle.name)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "description",
                "style_type",
                "face_lighting_model",
                "face_lighting_quality",
                "face_color_mode",
                "face_modifiers",
                "face_opacity_level",
                "face_specular_level",
                "color1",
                "color2",
                "face_style_mono_color",
                "edge_style_model",
                "edge_style",
                "edge_intersection_color",
                "edge_obscured_color",
                "edge_obscured_linetype",
                "edge_intersection_linetype",
                "edge_crease_angle",
                "edge_modifiers",
                "edge_color",
                "edge_opacity_level",
                "edge_width",
                "edge_overhang",
                "edge_jitter",
                "edge_silhouette_color",
                "edge_silhouette_width",
                "edge_halo_gap",
                "edge_isoline_count",
                "edge_hide_precision",
                "edge_style_apply",
                "style_display_settings",
                "brightness",
                "shadow_type",
                "unknown1",
                "internal_use_only_flag",
            ],
        )
        if self.acad_xdata:
            tagwriter.write_tags(self.acad_xdata)



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest

from ezdxf.entities import LWPolyline, Ellipse
from ezdxf import path


@pytest.fixture
def lwpolyline1():
    e = LWPolyline.new(
        dxfattribs={
            "layer": "0",
            "linetype": "Continuous",
            "color": 0,
            "flags": 0,
            "const_width": 0.0,
            "elevation": 2.999999999999999e99,
            "extrusion": (0.0, 0.0, 1.0),
        },
    )
    e.set_points(
        [
            (297888, 108770, 0.0, 0.0, 0.0512534542487669),
            (297930, 108335, 0.0, 0.0, 0.0),
        ]
    )
    return e


MAX_SAGITTA = 0.01


def test_flattening_by_arc(lwpolyline1: LWPolyline):
    entities = list(lwpolyline1.virtual_entities())
    for e in entities:
        if e.dxftype() == "ARC":
            points = list(e.flattening(MAX_SAGITTA))
            assert len(points) > 0

            ellipse = Ellipse.from_arc(e)
            points = list(ellipse.flattening(0.01))
            assert len(points) > 0


def test_flattening_by_ellipse(lwpolyline1: LWPolyline):
    entities = list(lwpolyline1.virtual_entities())
    for e in entities:
        if e.dxftype() == "ARC":
            ellipse = Ellipse.from_arc(e)
            points = list(ellipse.flattening(MAX_SAGITTA))
            assert len(points) > 0


def test_flattening_by_path(lwpolyline1):
    p = path.make_path(lwpolyline1)
    points = list(p.flattening(distance=MAX_SAGITTA))
    assert len(points) > 0


@pytest.fixture
def ellipse1():
    return Ellipse.new(
        dxfattribs={
            "center": (
                298998.9237372455,
                105908.98587737791,
                0.0218015606544459,
            ),
            "major_axis": (-2.005925505480262, 9.590684825374422e-13, 0.0),
            "ratio": 0.6052078628034204,
            "start_param": 7.898e-13,
            "end_param": 7.878142582740111e-13,
            "extrusion": (0.0, -0.0, 1.0),
        },
    )


def test_flattening_ellipse(ellipse1: Ellipse):
    points = list(ellipse1.flattening(MAX_SAGITTA))
    assert len(points) > 0


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
import ezdxf
from ezdxf.entities import get_font_name, DXFEntity
from ezdxf.lldxf import const


@pytest.fixture(scope="module")
def msp():
    doc = ezdxf.new()
    doc.styles.add("ARIAL", font="arial.ttf")
    return doc.modelspace()


def test_get_font_name_for_entity_without_font_support():
    e = DXFEntity()
    assert get_font_name(e) == const.DEFAULT_TEXT_FONT, "should return the default font"


def test_get_font_name_for_text(msp):
    text = msp.add_text("abc", dxfattribs={"style": "ARIAL"})
    assert get_font_name(text) == "arial.ttf"


def test_undefined_text_style(msp):
    text = msp.add_text("abc", dxfattribs={"style": "UNDEFINED"})
    assert get_font_name(text) == const.DEFAULT_TEXT_FONT


def test_get_font_name_for_mtext(msp):
    mtext = msp.add_mtext("abc", dxfattribs={"style": "ARIAL"})
    assert get_font_name(mtext) == "arial.ttf"


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
from typing import Sequence


def group_chars(s: str, n: int = 4, sep="-") -> str:
    chars = []
    for index, char in enumerate(reversed(s)):
        if index % n == 0:
            chars.append(sep)
        chars.append(char)
    if chars:
        chars.reverse()
        chars.pop()
        return "".join(chars)
    return ""


def bitmask_strings(
    value: int, base: int = 10, sep: str = "-"
) -> Sequence[str]:
    if base == 10:
        top, bottom = (
            "3322-2222-2222-1111-1111-1100-0000-0000",
            "1098-7654-3210-9876-5432-1098-7654-3210",
        )
    elif base == 16:
        top, bottom = (
            "1111-1111-1111-1111-0000-0000-0000-0000",
            "FEDC-BA98-7654-3210-FEDC-BA98-7654-3210",
        )
    else:
        raise ValueError(f"invalid base {base}, valid bases: 10, 16")
    top = top.replace("-", sep)
    bottom = bottom.replace("-", sep)
    l0 = len(top)
    bin_str = group_chars(bin(value)[2:], n=4, sep=sep)
    l1 = len(bin_str)
    return [
        top[l0 - l1 :],
        bottom[l0 - l1 :],
        bin_str,
    ]


def print_bitmask(value: int, *, base=10, sep="-"):
    lines = bitmask_strings(value, base, sep)
    assert len(lines) > 2
    divider_line = "=" * (max(map(len, lines)) + 4)
    print(divider_line)
    print("x0 :" + lines[0])
    print("0x :" + lines[1])
    print(divider_line)
    print("bin:" + lines[2])



# Copyright (c) 2011-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import (
    Any,
    Iterable,
    Iterator,
    KeysView,
    Optional,
    Sequence,
    TYPE_CHECKING,
    Union,
)
import logging
from collections import OrderedDict

from ezdxf.lldxf import const
from ezdxf.lldxf.tags import group_tags, Tags, DXFTag
from ezdxf.lldxf.types import strtag
from ezdxf.lldxf.validator import header_validator
from ezdxf.sections.headervars import (
    HEADER_VAR_MAP,
    version_specific_group_code,
)

if TYPE_CHECKING:
    from ezdxf.lldxf.tagwriter import AbstractTagWriter

logger = logging.getLogger("ezdxf")

MIN_HEADER_TEXT = """  0
SECTION
  2
HEADER
  9
$ACADVER
  1
AC1009
  9
$DWGCODEPAGE
  3
ANSI_1252
  9
$HANDSEED
  5
FF
"""

# Additional variables may be stored as DICTIONARYVAR in the OBJECTS
# section in the DICTIONARY "AcDbVariableDictionary" of the root dict.
# - CANNOSCALE
# - CENTEREXE
# - CENTERLTYPEFILE
# - CETRANSPARENCY
# - CMLEADERSTYLE
# - CTABLESTYLE
# - CVIEWDETAILSTYLE
# - CVIEWSECTIONSTYLE
# - LAYEREVAL
# - LAYERNOTIFY
# - LIGHTINGUNITS
# - MSLTSCALE


class CustomVars:
    """The :class:`CustomVars` class stores custom properties in the DXF header as
    $CUSTOMPROPERTYTAG and $CUSTOMPROPERTY values. Custom properties require DXF R2004
    or later, `ezdxf` can create custom properties for older DXF versions as well, but
    AutoCAD will not show that properties.

    """

    def __init__(self) -> None:
        self.properties: list[tuple[str, str]] = list()

    def __len__(self) -> int:
        """Count of custom properties."""
        return len(self.properties)

    def __iter__(self) -> Iterator[tuple[str, str]]:
        """Iterate over all custom properties as ``(tag, value)`` tuples."""
        return iter(self.properties)

    def clear(self) -> None:
        """Remove all custom properties."""
        self.properties.clear()

    def append(self, tag: str, value: str) -> None:
        """Add custom property as ``(tag, value)`` tuple."""
        # custom properties always stored as strings
        self.properties.append((tag, str(value)))

    def get(self, tag: str, default: Optional[str] = None):
        """Returns the value of the first custom property `tag`."""
        for key, value in self.properties:
            if key == tag:
                return value
        else:
            return default

    def has_tag(self, tag: str) -> bool:
        """Returns ``True`` if custom property `tag` exist."""
        return self.get(tag) is not None

    def remove(self, tag: str, all: bool = False) -> None:
        """Removes the first occurrence of custom property `tag`, removes all
        occurrences if `all` is ``True``.

        Raises `:class:`DXFValueError` if `tag`  does not exist.

        """
        found_tag = False
        for item in self.properties:
            if item[0] == tag:
                self.properties.remove(item)
                found_tag = True
                if not all:
                    return
        if not found_tag:
            raise const.DXFValueError(f"Tag '{tag}' does not exist")

    def replace(self, tag: str, value: str) -> None:
        """Replaces the value of the first custom property `tag` by a new
        `value`.

        Raises :class:`DXFValueError` if `tag`  does not exist.

        """
        properties = self.properties
        for index in range(len(properties)):
            name = properties[index][0]
            if name == tag:
                properties[index] = (name, value)
                return

        raise const.DXFValueError(f"Tag '{tag}' does not exist")

    def write(self, tagwriter: AbstractTagWriter) -> None:
        """Export custom properties as DXF tags. (internal API)"""
        for tag, value in self.properties:
            s = f"  9\n$CUSTOMPROPERTYTAG\n  1\n{tag}\n  9\n$CUSTOMPROPERTY\n  1\n{value}\n"
            tagwriter.write_str(s)


def default_vars() -> OrderedDict:
    vars = OrderedDict()
    for vardef in HEADER_VAR_MAP.values():
        vars[vardef.name] = HeaderVar(DXFTag(vardef.code, vardef.default))
    return vars


class HeaderSection:
    MIN_HEADER_TAGS = Tags.from_text(MIN_HEADER_TEXT)
    name = "HEADER"

    def __init__(self) -> None:
        self.hdrvars: dict[str, HeaderVar] = OrderedDict()
        self.custom_vars = CustomVars()

    @classmethod
    def load(cls, tags: Optional[Iterable[DXFTag]] = None) -> HeaderSection:
        """Constructor to generate header variables loaded from DXF files
        (untrusted environment).

        Args:
            tags: DXF tags as Tags() or ExtendedTags()

        (internal API)
        """
        if tags is None:  # create default header
            # files without a header have the default version: R12
            return cls.new(dxfversion=const.DXF12)
        section = cls()
        section.load_tags(iter(tags))
        return section

    @classmethod
    def new(cls, dxfversion=const.LATEST_DXF_VERSION) -> HeaderSection:
        section = HeaderSection()
        section.hdrvars = default_vars()
        section["$ACADVER"] = dxfversion
        return section

    def load_tags(self, tags: Iterable[DXFTag]) -> None:
        """Constructor to generate header variables loaded from DXF files
        (untrusted environment).

        Args:
            tags: DXF tags as Tags() or ExtendedTags()

        """
        _tags = iter(tags) or iter(self.MIN_HEADER_TAGS)
        section_tag = next(_tags)
        name_tag = next(_tags)

        if section_tag != (0, "SECTION") or name_tag != (2, "HEADER"):
            raise const.DXFStructureError("Critical structure error in HEADER section.")

        groups = group_tags(header_validator(_tags), splitcode=9)
        custom_property_stack = []  # collect $CUSTOMPROPERTY/TAG
        for group in groups:
            name = group[0].value
            value = group[1]
            if name in ("$CUSTOMPROPERTYTAG", "$CUSTOMPROPERTY"):
                custom_property_stack.append(value.value)
            else:
                self.hdrvars[name] = HeaderVar(value)

        custom_property_stack.reverse()
        while len(custom_property_stack):
            try:
                self.custom_vars.append(
                    tag=custom_property_stack.pop(),
                    value=custom_property_stack.pop(),
                )
            except IndexError:  # internal exception
                break

    @classmethod
    def from_text(cls, text: str) -> HeaderSection:
        """Load constructor from text for testing"""
        return cls.load(Tags.from_text(text))

    def _headervar_factory(self, key: str, value: Any) -> DXFTag:
        if key in HEADER_VAR_MAP:
            factory = HEADER_VAR_MAP[key].factory
            return factory(value)
        else:
            raise const.DXFKeyError(f"Invalid header variable {key}.")

    def __len__(self) -> int:
        """Returns count of header variables."""
        return len(self.hdrvars)

    def __contains__(self, key) -> bool:
        """Returns ``True`` if header variable `key` exist."""
        return key in self.hdrvars

    def varnames(self) -> KeysView:
        """Returns an iterable of all header variable names."""
        return self.hdrvars.keys()

    def export_dxf(self, tagwriter: AbstractTagWriter) -> None:
        """Exports header section as DXF tags. (internal API)"""

        def _write(name: str, value: Any) -> None:
            if value.value is None:
                logger.info(f"did not write header var {name}, value is None.")
                return
            tagwriter.write_tag2(9, name)
            group_code = version_specific_group_code(name, dxfversion)
            # fix invalid group codes
            if group_code != value.code:
                value = HeaderVar((group_code, value.value))
            tagwriter.write_str(str(value))

        dxfversion: str = tagwriter.dxfversion
        write_handles = tagwriter.write_handles

        tagwriter.write_str("  0\nSECTION\n  2\nHEADER\n")
        save = self["$ACADVER"]
        self["$ACADVER"] = dxfversion
        for name, value in header_vars_by_priority(self.hdrvars, dxfversion):
            if not write_handles and name == "$HANDSEED":
                continue  # skip $HANDSEED
            _write(name, value)
            if name == "$LASTSAVEDBY":  # ugly hack, but necessary for AutoCAD
                self.custom_vars.write(tagwriter)
        self["$ACADVER"] = save
        tagwriter.write_str("  0\nENDSEC\n")

    def get(self, key: str, default: Any = None) -> Any:
        """Returns value of header variable `key` if exist, else the `default`
        value.

        """
        if key in self.hdrvars:
            return self.__getitem__(key)
        else:
            return default

    def __getitem__(self, key: str) -> Any:
        """Get header variable `key` by index operator like:
        :code:`drawing.header['$ACADVER']`

        """
        try:
            return self.hdrvars[key].value
        except KeyError:  # map exception
            raise const.DXFKeyError(str(key))

    def __setitem__(self, key: str, value: Any) -> None:
        """Set header variable `key` to `value` by index operator like:
        :code:`drawing.header['$ANGDIR'] = 1`

        """
        try:
            tags = self._headervar_factory(key, value)
        except (IndexError, ValueError):
            raise const.DXFValueError(str(value))
        self.hdrvars[key] = HeaderVar(tags)

    def __delitem__(self, key: str) -> None:
        """Delete header variable `key` by index operator like:
        :code:`del drawing.header['$ANGDIR']`

        """
        try:
            del self.hdrvars[key]
        except KeyError:  # map exception
            raise const.DXFKeyError(str(key))

    def reset_wcs(self):
        """Reset the current UCS settings to the :ref:`WCS`."""
        self["$UCSBASE"] = ""
        self["$UCSNAME"] = ""
        self["$UCSORG"] = (0, 0, 0)
        self["$UCSXDIR"] = (1, 0, 0)
        self["$UCSYDIR"] = (0, 1, 0)
        self["$UCSORTHOREF"] = ""
        self["$UCSORTHOVIEW"] = 0
        self["$UCSORGTOP"] = (0, 0, 0)
        self["$UCSORGBOTTOM"] = (0, 0, 0)
        self["$UCSORGLEFT"] = (0, 0, 0)
        self["$UCSORGRIGHT"] = (0, 0, 0)
        self["$UCSORGFRONT"] = (0, 0, 0)
        self["$UCSORGBACK"] = (0, 0, 0)


def header_vars_by_priority(
    header_vars: dict[str, HeaderVar], dxfversion: str
) -> Iterable[tuple]:
    order = []
    for name, value in header_vars.items():
        vardef = HEADER_VAR_MAP.get(name, None)
        if vardef is None:
            logger.info(f"Header variable {name} ignored, dxfversion={dxfversion}.")
            continue
        if vardef.mindxf <= dxfversion <= vardef.maxdxf:
            order.append((vardef.priority, (name, value)))
    order.sort()
    for priority, tag in order:
        yield tag


class HeaderVar:
    def __init__(self, tag: Union[DXFTag, Sequence]):
        self.tag = tag

    @property
    def code(self) -> int:
        return self.tag[0]

    @property
    def value(self) -> Any:
        return self.tag[1]

    @property
    def ispoint(self) -> bool:
        return self.code == 10

    def __str__(self) -> str:
        if self.ispoint:
            code, value = self.tag
            s = []
            for coord in value:
                s.append(strtag((code, coord)))
                code += 10
            return "".join(s)
        else:
            return strtag(self.tag)  # type: ignore
